(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Omega Program ClassicalChoice.
Require Import ssr vgtac basis myclassical model_common model model_properties gw.
Require Import lang_common lang_src_lsi lang_tgt lang_spec util ts.

Unset Implicit Arguments.       
Global Set Bullet Behavior "Strict Subproofs".


Notation AWK := 0.              

Opaque pmplus.

(** * Source module. *)

Local Notation X := 0.
Local Notation Y := 1.
Local Notation F := 2.

Definition vs_inner x :=
  Vabs F (Elet Y (Ewrite x (Enat 1)) (
          Elet Y (Eapp (Evar F) Eunit) (
          Eread x))).

Definition vs: val src_lsi :=
  Vabs X (Elet X (Enew (Enat 0)) (vs_inner (Evar X))).

Definition ms: module src_lsi :=
  [(AWK, vs)].

Definition Tawk := FTarrow Tunit (Tarrow (Tarrow Tunit Tunit) Tnat).



(** * Target module. *)


Definition mt_code_tbe :=
  enc (i_jnz r_aux (o_reg r_clo)) :: (** 17 *)

  
  enc (i_ld r_aux (o_imm (enc (i_opr Plus r_aux (o_reg r_j) (o_imm 23))))) :: (** 18 *)
  enc (i_opr Plus r_i (o_reg r_j) (o_imm 8)) ::
  enc (i_sto (o_hind r_i o_zero) r_aux) ::
  enc (i_ld r_aux (o_imm (enc (i_jmp (o_reg r_aux))))) ::
  enc (i_sto (o_hind r_i o_one) r_aux) ::

  
  enc (i_sto (o_sind r_sp o_zero) r_ret) :: (** 23 *)
  enc (i_opr Plus r_sp (o_reg r_sp) (o_imm 1)) ::
  enc (i_opr Plus r_ret (o_reg r_j) (o_imm 28)) ::

  
  enc (i_sto (o_reg r_clo) r_arg) ::
  enc (i_jmp (o_hind r_clo o_zero)) ::

  
  enc (i_opr Minus r_sp (o_reg r_sp) (o_imm 1)) :: (** 28 *)
  enc (i_ld r_ret (o_sind r_sp o_zero)) ::
  enc (i_ld r_arg (o_imm 1)) ::
  enc (i_jmp (o_reg r_ret)) ::
  nil.                          (** 32 *)


Definition encrypt_from n ws :=
  (** First [n] stay decrypted. *)
  firstn n ws ++ map (fun w => w + 666) (skipn n ws).


Definition mt_code' (patched: bool) (ndecrypted: nat) : list word :=

  (** create and return closure *)
  enc (i_lpc r_j) :: (** 0 *)
  enc (i_ld r_arg (o_imm 1)) ::
  enc (i_alc r_arg r_arg) ::
  enc (i_opr Plus r_aux (o_reg r_j) (o_imm 6)) ::
  enc (i_sto (o_hind r_arg o_zero) r_aux) ::
  enc (i_jmp (o_reg r_ret)) ::

  (** closure begin *)
  enc (i_lpc r_j) :: (** 6 *)
  enc (i_opr Minus r_j (o_reg r_j) (o_imm 6)) ::
  enc (i_opr Plus r_aux (o_reg r_j) (o_imm (if patched then 23 else 17))) ::
  enc ((if patched then i_jmp else i_ld r_i) (o_reg r_aux)) ::

  (** decrypt in loop *)
  enc (i_ld r_aux (o_hind r_i o_zero)) :: (** 10 *)
  enc (i_opr Minus r_aux (o_reg r_aux) (o_imm 666)) ::
  enc (i_sto (o_hind r_i o_zero) r_aux) ::
  enc (i_opr Plus r_i (o_reg r_i) (o_imm 1)) ::
  enc (i_opr Plus r_aux (o_reg r_j) (o_imm 32)) ::
  enc (i_opr Minus r_aux (o_reg r_aux) (o_reg r_i)) ::
  enc (i_opr Plus r_clo (o_reg r_j) (o_imm 10)) :: (** 16 *)
  encrypt_from ndecrypted mt_code_tbe.


Definition mt_code (encrypted: bool) : list word :=
  mt_code' (negb encrypted)
           (if encrypted then 0 else length mt_code_tbe).


Definition mt: module tgt_lsi :=
  [ {| g_cptrs := [(AWK, 0)]; g_data := mt_code true |} ].



(** * Well-formedness. *)

Lemma ms_wf: src_wfmod ms [AWK].
Proof.
  rr; auto.
Qed.

Lemma mt_wf: tgt_wfmod mt [AWK].
Proof.
  rr; splits.
  - s; i; des; vauto.
  - s; i; des; vauto; ss; des; vauto; omega.
Qed.



(** * Correctness proof. *)


(** ** Miscellaneous properties about the tgt instructions. *)


Lemma mt_code_valid: forall l b hp
    (VLD: pmvalid ({l |=> mt_code b} ## hp)),
  pmvalid ({l |=> mt_code false} ## hp).
Proof.
  i; eapply pmvalid_plus_cells; eauto.
  s; repeat f_equal; destruct b; auto.
Qed.

Lemma modify_code: forall l n
    (LEN: n < length mt_code_tbe),
  exists h w,
    {l |=> mt_code' false n} = {l + 17 + n |-> w + 666} ## h /\
    {l |=> mt_code' false (S n)} = {l + 17 + n |-> w} ## h.
  (** Encrypted code starts at l + 17 + n. *)
Proof.
  i; unfold mt_code'.
  rewrite ->!pmcells_cons.
  rhs rewrite ->!pmplus_assoc, (pmplus_comm _ {_|=>_}), <-!pmplus_assoc.
  unfold encrypt_from; rewrite -(plus_comm 17); simpl (17 + _).
  rewrite pmcells_app firstn_length.
  rhs rewrite -pmplus_assoc pmplus_swap.
  simpl length; simpl in LEN; rewrite min_l; [|omega].
  rewrite -(firstn_skipn n mt_code_tbe).
  assert (exists h t, skipn n mt_code_tbe = h::t).
  { unfold mt_code_tbe.
    do 16 (destruct n; s; eauto); omega.
  }
  des; rewrite H; simpl nth.
  rewrite skipn_app.
  rewrite firstn_length min_l; [|s; omega].
  rewrite minus_diag skipn_eq_nil
  ; last by rewrite firstn_length min_l; [|s; omega].
  simpl map; rewrite pmcells_cons -pmplus_assoc -H.
  replace h with (hd 42 (skipn n mt_code_tbe)); [|by rewrite H].
  replace t with (tl (skipn n mt_code_tbe)); [|by rewrite H].
  esplits; eauto; []; pmcells_fold.
  rhs rewrite pmplus_comm -pmplus_assoc pmplus_swap.
  rewrite !pmcells_cons -!pmplus_assoc; repeat f_equal.
  rewrite pmcells_nil; noemp; rewrite firstn_skipn.
  pmcells_fold.
  rewrite H; simpl hd; simpl tl.
  erewrite firstn_S; [|by auto]; rewrite H; simpl hd.
  rewrite -app_assoc pmcells_app; f_equal.
  rewrite firstn_length min_l; [|s; omega].
  f_equal; s; f_equal; unfold mt_code_tbe in H; f_equal.
  assert (E: skipn 1 (h :: t) = t) by auto.
  rewrite -H skipn_skipn in E; ss; auto.
Grab Existential Variables.
  exact 0.
Qed.

Lemma mt_code'_length: forall n
    (LEN: n < length mt_code_tbe),
  length (mt_code' false n) = length (mt_code' false (S n)).
Proof.
  i; ss; repeat f_equal; unfold encrypt_from.
  rewrite !app_length !map_length !skipn_length.
  rewrite !firstn_length !min_l; try (s; omega).
  simpl length; omega.
Qed.



(** ** Local world. *)

Record mylwstate :=
  { laddr: word                 (** tgt *)
  ; encrypted: bool             (** tgt *)
  ; retaddrs: stack             (** tgt *)
  ; instances: list (location * bool) (** src *)
  }.



Definition switch (s: mylwstate) l sp retaddr : mylwstate :=
  {| laddr := laddr s
   ; encrypted := false
   ; retaddrs := {sp |-> retaddr} ## retaddrs s
   ; instances := map (fun p => if eqdec (fst p) l then (l, true) else p) (instances s)
  |}.















  







Definition myext (s' s: mylwstate) :=
  laddr s' = laddr s /\
  <<ENC: Bool.leb (encrypted s') (encrypted s)>> /\
  <<INS: forall l b (IN: In (l, b) (instances s)),
           exists b', <<IN: In (l, b') (instances s')>> /\
                      <<CAL: Bool.leb b b'>> >>.

Definition myextpub (s' s: mylwstate) :=
  <<EXT: myext s' s>> /\
  <<ST: retaddrs s = retaddrs s'>>.

Lemma myext_refl: forall s,
  myext s s.
Proof.
  i; rr; splits; eauto.
Qed.
Hint Resolve myext_refl.

Lemma myext_trans: forall s s' s''
    (EXT': myext s'' s')
    (EXT: myext s' s),
  myext s'' s.
Proof.
  unfold myext; i; des; esplits.
  - congruence.
  - destruct (encrypted s), (encrypted s'), (encrypted s''); ss; eauto.
  - i; apply INS in IN; des; apply INS0 in IN0; des.
    esplits; eauto using Bool_leb_trans.
Qed.

Lemma myextpub_trans: forall s s' s''
    (EXT': myextpub s'' s')
    (EXT: myextpub s' s),
  myextpub s'' s.
Proof.
  unfold myextpub; i; des; esplits; eauto using myext_trans; congruence.
Qed.

Program Definition mylwsts: transys :=
  {| state := mylwstate
   ; ext := myext
   ; extpub := myextpub
  |}.
Next Obligation. eauto using myext_trans. Qed.
Next Obligation. split; [rr; esplits|]; eauto. Qed.
Next Obligation. eauto using myextpub_trans. Qed.
Next Obligation. rrdes EXTP; auto. Qed.

Program Definition mylwconf: method_conf tgt_lsi src_lsi (tsprod (gwsts ts_gw) mylwsts) :=
  {| confrel g s ca cb :=
       exists ha hb,
         ca = {| cpc := Nothing; crf := Nothing; cst := retaddrs (snd s); chp := ha |} /\
         cb = (hb, Nothing, Nothing) /\
         ha = {laddr (snd s) |=> mt_code (encrypted (snd s))} ## {laddr (snd s) - 2 |-> laddr (snd s) - 1} /\
         <<INS: forall lb called (IN: In (lb, called) (instances (snd s))),
                  pmread hb lb (if called then (Vnat 1) else (Vnat 0))>>
  |}.
Next Obligation.
  s; esplits; eauto.
Qed.                            

Program Definition mylwname: method_name tgt_lsi src_lsi (tsprod (gwsts ts_gw) mylwsts) :=
  {| supp := fun _ => False
   ; name := fun _ _ _ _ _ => False
  |}.

Program Definition mylw: world_loca ts_gw :=
  {| lwwfo := gwfo
   ; lwsts := mylwsts; lwconf := mylwconf; lwname := mylwname |}.
Next Obligation.
  unfold tgt_extra, src_extra; unnw; subst; ss; eauto.
Qed.

Lemma mylw_stable: stable mylw.
Proof.
  ii; rrdes CR; rr; esplits; [by eauto|].
  rr; esplits; eauto.
Qed.

Lemma mylw_confrel_shape: forall g s ca cb
    (CR: confrel mylw g s ca cb),
  exists ha hb,
    ca = {| cpc := Nothing; crf := Nothing; cst := retaddrs (snd s); chp := ha |} /\
    cb = (hb, Nothing, Nothing).
Proof.
  i; destruct ca, cb; rrdes CR; ss; subst; eauto.
Qed.

Lemma switch_ext: forall s x sp retaddr,
  ext mylw (switch s x sp retaddr) s.
Proof.
  i; destruct s; econs; s; splits; auto.
  i; exists (if eq_nat_dec l x then true else b); split.
  - eapply in_map_iff; exists (l, b); s; des_if; subst; eauto.
  - des_if; rr; des_if; eauto.
Qed.


(** ** Tactics. *)


Ltac ts_prepare' :=
  intro HR;                     
  match type of HR with confrel _ _ _ ?csa _ =>
    eapply pesim_real; intros ma ? RLa ?;
    destruct csa as [cpc ? ? ?], ma;
    destruct_tgt_real RLa;
    destruct cpc; [by vauto|clarify|by vauto];
    rrdes HR WHX WHX0; subst;
    hexploit tgt_confpred_shape; eauto;
    hexploit src_confpred_shape; eauto;
    hexploit ts_refconfrel_shape; eauto;
    hexploit mylw_confrel_shape; eauto;
    i; des; clarify; noemp;
    simpl cplus in *; clarify; noemp; clean;
    rewrite <-!pmplus_assoc in *;
    rrdes WHY; clarify; clean
  end.

Ltac ts_prepare :=
  let HR := fresh in
  intros ? ? HR; revert HR; ts_prepare'.


(** ** Main proof. *)

Lemma awkward_encrypt: forall i j ka kb g si s l mrf st hp es
    (SIM: forall v_clo v_aux v_i,
            esim (wlift mylw) j ka kb g si s false Tnat
               {| cpc := Just (l + 23)
                ; crf := Just [v_clo//r_clo]>>[v_aux//r_aux]>>
                              [v_i//r_i]>>[l//r_j]>>mrf
                ; cst := st
                ; chp := {l |=> mt_code false} ## hp |}
               es)
    (GLE: wfo_le_n (4 + (8 * length mt_code_tbe + 5)) j i),
  esim (wlift mylw) i ka kb g si s false Tnat
       {| cpc := Just (l + 6)
        ; crf := Just mrf
        ; cst := st
        ; chp := {l |=> mt_code true} ## hp |}
       es.
Proof.
  i; eapply pesim_real; intros ma ? RLa ?.
  destruct ma; destruct_tgt_real RLa.

  
  inversion RF; inversion PC; subst; clear PC RF.

  eapply pesim_dance; first (by eauto); clear RLa0 SV HV FIN.
  { rewrite plus_comm.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Slpc].
      eapply pmread_plus; eauto.
      unfold mt_code, mt_code', negb; rewrite ->!pmcells_cons; pmplus_eq''.
    }
    { depmcplus0. }
    
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - eapply pmread_plus; eauto.
        unfold mt_code, mt_code', negb; rewrite ->!pmcells_cons; pmplus_eq''.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }
    
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - eapply pmread_plus; eauto.
        unfold mt_code, mt_code', negb; rewrite ->!pmcells_cons; pmplus_eq''.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }
    
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        unfold mt_code, mt_code', negb; rewrite ->!pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus0. }
    eapply dance_O; instantiate (1 := fun _ _ _ => depmach0); eauto.
  }

  (** Setting up the induction.
     [m] is the number of remaining encrypted instructions.
  *)
  i; unfold dmach_apply, mt_code; simpl negb.
  eapply (wfo_le_n_plus _ 4) in GLE; [|by rewrite plus_comm; eauto]; des.
  esplits; eauto; [].
  i; esplits; eauto; [].
  clear dependent cfas; clear dependent i; depgen ea'.
  rewrite -minus_n_O; generalize (l + 17) at 2.
  rename mrf0 into mrf.
  rewrite (tmext_self r_clo mrf); generalize (mrf r_clo).
  set (len := length mt_code_tbe) in *; pose (m := len).
  match goal with
    |- context[{|mpc := _; mrf := [?x//_]>>_; mst := _; mhp := _|}] =>
    replace x with (x + (len - m)); [|by auto]
  end.
  change 0 with (len - m) at 2; depgen ii; change len with m at 1.
  assert (LE: 1 <= m <= len) by (unfold m, len; s; omega).
  depgen m; induction m using (well_founded_ind lt_wf).

  i; rrdes XTRb RLb0; clarify.
  destruct es as [[]]; s in Ec; noemp; clarify.

  eapply pesim_dance; first (by eauto); clear RLa FIN.
  { hexploit (modify_code l (len - m))
    ; [unfold len; simpl length; omega|]; i; des.
    hide $ X X0.              

      
      

    Arguments minus _ _: simpl nomatch.
    
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - eapply pmread_plus; eauto.
        unfold mt_code, mt_code'; rewrite ->!pmcells_cons; pmplus_eq''.
      - econs; s; tmext_simpl.
        eapply pmread_plus; eauto.
        rewrite X -pmplus_assoc; eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          unfold mt_code, mt_code'; rewrite ->!pmcells_cons; pmplus_eq''.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }
      tmext_uniq.
      rewrite plus_comm; simpl (_ + _ - _); rewrite -minus_n_O.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Ssto].
        - eapply pmread_plus; eauto.
          unfold mt_code, mt_code'; rewrite ->!pmcells_cons; pmplus_eq''.
        - set (mrf' := [w//r_aux]>>[l + 17 + (len - m)//r_i]>>[l//r_j]>>[n//r_clo]>>mrf).
          unfold mrf' at 2; tmext_simpl; rewrite X -pmplus_assoc.
          prep_store; eauto.
      }
      { tmext_simpl; rewrite pmplus_assoc -X0.
        depmcplus0.
        - eapply pmvalid_plus_cells; eauto using mt_code'_length.
          eapply mt_code'_length; unfold len; s; omega.
        - eapply pmfinite_plus_cells; eauto.
      }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          unfold mt_code, mt_code'; rewrite ->!pmcells_cons; pmplus_eq''.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          unfold mt_code, mt_code'; rewrite ->!pmcells_cons; pmplus_eq''.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          unfold mt_code, mt_code'; rewrite ->!pmcells_cons; pmplus_eq''.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          unfold mt_code, mt_code'; rewrite ->!pmcells_cons.
          unfold encrypt_from; s; rewrite pmcells_cons; pmplus_eq'.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }

      eapply dance_O; instantiate (1 := fun _ _ _ _ _ _ => depmach0); eauto.
    }

    i; rrdes XTRb RLb0; clarify.
    s in Ec; noemp; clarify.
    eapply (wfo_le_n_plus (8 * (m-1) + 6) 7) in GLEa; [|omega]; des.
    esplits; first 1 last.
    + right; split; eauto.
    + inst_pairs; s; noemp; rr; eauto 10.
    + i; frames_nodep RLa cfas; destruct ea'0; clear FIN.
      destruct_tgt_real RLa; esplits; [by eauto|].

      destruct (LEM (m = 1)).
      { (*| This is the last loop iteration. *)

        eapply pesim_dance; first (by eauto); clear FIN SV HV.
        { 
          dance_init.
          { eapply Sok; [s; i; omega|eapply Sjz].
            - eapply pmread_plus; eauto.
              unfold mt_code; rewrite !pmcells_cons; pmplus_eq''.
            - tmext_simpl.
              unfold len; s; omega.
          }
          { depmcplus0. }

          dance_init.
          { eapply Sok; [s; i; omega|eapply Sld].
            - eapply pmread_plus; eauto.
              unfold mt_code; rewrite !pmcells_cons; pmplus_eq''.
            - eauto.
          }
          { depmcplus0. }
          tmext_uniq.

          dance_init.
          { eapply Sok; [s; i; omega|eapply Sopr].
            - eapply pmread_plus; eauto.
              unfold mt_code; rewrite !pmcells_cons; pmplus_eq''.
            - eauto.
            - eauto.
            - eauto.
          }
          { depmcplus0. }
          tmext_uniq.

          dance_init.
          { eapply Sok; [s; i; omega|eapply Ssto].
            - eapply pmread_plus; eauto.
              unfold mt_code; rewrite !pmcells_cons; pmplus_eq''.
            - unfold len in *; simpl length in *.
              tmext_simpl; simpl minus; s in mrf.
              rewrite 9!pmcells_cons; rewrite <-!pmplus_assoc.
              rewrite -8!(pmplus_swap {S (S (S (S (S (S (S (S l)))))))|->_}).
              prep_store; eauto.              
          }
          { s; tmext_simpl.
            rewrite 8!(pmplus_swap {l + 8 |-> _}) -(plus_comm 8); s.
            pmcells_fold.     
            set (xxx:=enc (i_lpc _) :: enc (i_ld _ _) :: _).
            instantiate (1 := depmach0).
            eapply tgt_mcplus; first (by s; eauto).
            rr; esplits; s; tmext_simpl.
            - eauto.
            - eauto.
            - eauto.
            - eauto.
            - rewrite pmplus_assoc; eauto.
            - eapply pmvalid_plus_cells; eauto. 
            - eapply pmfinite_plus_cells; eauto.
          }

          dance_init.
          { eapply Sok; [s; i; omega|eapply Sld].
            - eapply pmread_plus; eauto.
              unfold mt_code; rewrite !pmcells_cons; pmplus_eq''.
            - eauto.
          }
          { depmcplus0. }
          tmext_uniq.

          dance_init.
          { eapply Sok; [s; i; omega|eapply Ssto].
            - eapply pmread_plus; eauto.
              rewrite -(firstn_skipn 22 (enc _ :: _)) pmcells_app; s.
              rewrite (pmcells_cons (l + 22)).
              rhs rewrite -pmplus_assoc pmplus_swap plus_comm.
              pmplus_eq'.
            - tmext_simpl.
              rewrite -(firstn_skipn 9 (enc _ :: _)) pmcells_app; s.
              rewrite (pmcells_cons (l + 9)).
              replace (l + 9) with (add_offset (8 + l) o_one); [|s; omega].
              rewrite <-!pmplus_assoc, pmplus_swap.
              prep_store; eauto.
          }
          { s; tmext_simpl.
            rewrite pmplus_swap.
            rewrite (pmplus_assoc {_ |-> _}) -pmcells_cons.
            replace (S (S (S (S (S (S (S (S (S l))))))))) with (l + 9); [|omega].
            rewrite pmplus_assoc -pmcells_app'; s; [|omega].
            replace (enc _::_) with (mt_code false); [|by auto].
            
            instantiate (1 := depmach0).
            eapply tgt_mcplus; first (by s; eauto).
            rr; esplits; s; tmext_simpl.
            - eauto.
            - eauto.
            - eauto.
            - eauto.
            - rewrite pmplus_assoc; eauto.
            - eapply pmvalid_plus_cells; [exact HV0|by s; eauto]. 
            - eapply pmfinite_plus_cells; eauto.
          }

          eapply dance_O; instantiate (1 := fun _ _ _ _ _ => depmach0); eauto.
        }

        
        i; rrdes XTRb RLb0; clarify.
        s in Ec; noemp; clarify.
        esplits; first 2 last; last first.
        + right; split; eauto using wfo_le_n_add_S.
        + i; frames_nodep RLa cfas; destruct ea'0; clear FIN.
          destruct_tgt_real RLa; esplits; [by eauto|].
          replace (S _) with (l + 23); [|omega].
          repeat (rewrite -(@tmext_swap _ _ _ r_clo); [|by vauto]).
          rewrite tmext_shadow.
          apply SIM.
        + s; rr; eauto 10.
      }
      { (*| This is NOT the last loop iteration. *)

        eapply pesim_dance; first (by eauto); clear FIN SV HV.
        {

          assert (exists n, l + 32 - (l + 17 + (len - m) + 1) = S n).
          { destruct (l + 32 - (l + 17 + (len - m) + 1)) eqn:E; [|by eauto].
            exfalso; clear - E f LE LE0; unfold len in E; s in E.
            destruct m; [|destruct m; first by vauto]; omega.
          }
          des.

          dance_init.
          { eapply Sok; [s; i; omega|eapply Sjnz].
            - eapply pmread_plus; eauto.
              unfold mt_code; rewrite !pmcells_cons; pmplus_eq''.
            - tmext_simpl; eauto.
            - eauto.
          }
          { depmcplus0. }

          eapply dance_O; instantiate (1 := depmach0); eauto.
        }

        eapply (wfo_le_n_plus (8 * (m-1) + 5) 1) in GLEa0; [|omega]; des.
        i; esplits; eauto.
        i; frames_nodep RLa cfas; destruct ea'0.
        destruct_tgt_real RLa; esplits; [by eauto|].
        eapply (H (m-1)).
        - omega.
        - destruct m; [|destruct m; first by vauto]; omega.
        - eauto.
        - rewrite {2}plus_comm in RLa1; s in RLa1.
          replace (l + 17 + (len - (m - 1))) with (l + 17 + (len - m) + 1); [|omega].
          replace (len - (m - 1)) with (S (len - m)); [|omega].
          rhs rewrite <-tmext_swap; [|by vauto].
          clear RLa0.
          rewrite -(plus_comm 10).
          rrdes RLa1.
          rr; splits.
          + eauto.
          + s; repeat (rewrite (@tmext_swap _ _ _ r_clo); [|by vauto]).
            rewrite tmext_shadow; eauto.
          + eauto.
          + eauto.
          + s; reflexivity.        
          + s; s in HV0; exact HV0.
          + s; s in FIN0; exact FIN0.
      }
Opaque mt_code. Opaque mt_code'. Opaque mt_code_tbe.
Qed. (** These Opaque commands are for working around a bug in Coq. *)
Transparent mt_code. Transparent mt_code'. Transparent mt_code_tbe.


Lemma awkward_lsim_aux: forall g (s: state (wlift mylw)) vt x called
    (IN: In (x, called) (instances (snd s)))
    (LA: ts_db s (ts_fun (laddr (snd s) + 6)) vt)
    (RJ: ts_rf s r_j = laddr (snd s))
    (GK: gknow (wlift mylw) g),
  lsim (wlift mylw) g s (FTarrow (Tarrow Tunit Tunit) Tnat)
       vt (vs_inner (Eloc x)).
Proof.
  i; econs.
  { econs; rr; splits; auto; s; i; des; vauto. }

  i; invc STK; invc RF VR RK; rrdes CA; clear CB.
  invc APPa APPb; unfold vs in VAL; s in VAL; clarify; s; tmext_simpl.
  destruct s' as [[[[]]]]; clean; clear dependent v'.
  apply EXT in LA; clean.

  apply pesim_gknow; i.
  apply pesim_acquire.
  rename PC into CLO.             
  ts_prepare.

  hexploit tgt_vquery_fun_uniq; [by eauto|exact CLO|exact LA|]; i; subst mpc.

  dup EXT; rrdes EXT EXT1.
  dup RV; invc RV REL VR; rrdes CB; invc SVQ.
  hexploit INS0; eauto; []; i; des; clean.
  hexploit INS; eauto; []; i; rrdes X; subst.
  rewrite (pmplus_comm refha) (pmplus_swap hp) EXT1.

  assert (MAIN: forall i v_clo v_aux v_i,
    esim (wlift mylw) i (mrf r_ret) kb g' (mrf, P, s0, s1, s2)
         (mrf, P, s0, s1, s2) false Tnat
         {| cpc := Just (laddr (snd s) + 23);
            crf := Just [v_clo//r_clo]>>[v_aux//r_aux]>>[v_i//r_i]>>[laddr (snd s)//r_j]>>mrf;
            cst := freest ## retaddrs s2;
            chp := {laddr (snd s) |=> mt_code false} ## {laddr (snd s) - 2 |-> laddr (snd s) - 1} ## hp ## refha
         |}
         (refhb ## {x |-> if b' then Vnat 1 else Vnat 0} ## h', 
          Just s0,
          Just (app_cont kb (Eapp (Vabs Y (Eapp (Vabs Y (Eread (Eloc x)))
                                                (Eapp (Efix xf xa e) Eunit)))
                                  (Ewrite (Eloc x) (Enat 1))))
         )).
  { 
    
    i; eapply pesim_stepb.
    i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify. 
    esplits.
    { change (app_cont _ _)
        with (app_cont (SKappR (Vabs Y (Eapp (Vabs Y (Eread (Eloc x)))
                                             (Eapp (Efix xf xa e) Eunit))) :: kb)
                       (Ewrite (Eloc x) (Enat 1))).
      rewrite pmplus_swap -pmplus_assoc.
      inst_pairs; econs; eapply Rok, red_ctx; econs; econs.
    }
    right; split.
    { rewrite -2!pmplus_assoc pmplus_swap in PV0, FIN1.
      inst_pairs; s; noemp; rr; esplits.
      - eauto.
      - rewrite pmplus_assoc; eauto.
      - rewrite -!pmplus_assoc; eauto using pmvalid_plus_cell.
      - rewrite -!pmplus_assoc; eauto using pmfinite_plus_cell.
    }

    (** Expose next free stack cell. *)
    hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
    des; repeat (destruct vs0; [by vauto|]); destruct vs0; [|by vauto].
    rewrite <-pmcells_sng in *; clear NST.

    (** Expose closure in memory. *)
    dup CPa; rrdes CA CPa; dup CA; apply HP in CA; clear PC RF ST HP; clean.

    eapply pesim_dance_call.
    { instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto.
      + rewrite EXT1 (pmplus_comm refha) (pmplus_swap hp) in HV, FIN.
        eapply mt_code_valid.
        rewrite <-!pmplus_assoc in HV; eauto.
      + rewrite EXT1 (pmplus_comm refha) (pmplus_swap hp) in HV, FIN.
        rewrite <-!pmplus_assoc in HV, FIN.
        eapply pmfinite_plus_cells; eauto.
    }
    { clear RLa0 SV HV FIN FIN0.
      rename mrf into mrf0.
      rewrite plus_comm; s.


      dance_init.
      { eapply Sok; [s; i; omega|eapply Ssto].
        - eapply pmread_plus; eauto.
          unfold mt_code, mt_code'.
          rewrite -(firstn_skipn 23 (enc _ :: _)) pmcells_app; s.
          rewrite (pmcells_cons (_ + 23)).
          rhs rewrite -pmplus_assoc pmplus_swap plus_comm.
          pmplus_eq'.
        - prep_store; eauto.
      }
      { depmcplus0; eauto using pmvalid_plus_cell. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          unfold mt_code, mt_code'.
          rewrite -(firstn_skipn 24 (enc _ :: _)) pmcells_app; s.
          rewrite (pmcells_cons (_ + 24)).
          rhs rewrite -pmplus_assoc pmplus_swap plus_comm.
          pmplus_eq'.
        - eauto.
        - eauto.
        - eauto.
      }
      { tmext_simpl; depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          unfold mt_code, mt_code'.
          rewrite -(firstn_skipn 25 (enc _ :: _)) pmcells_app; s.
          rewrite (pmcells_cons (_ + 25)).
          rhs rewrite -pmplus_assoc pmplus_swap plus_comm.
          pmplus_eq'.
        - eauto.
        - eauto.
        - eauto.
      }
      { tmext_simpl; depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Ssto].
        - eapply pmread_plus; eauto.
          unfold mt_code, mt_code'.
          rewrite -(firstn_skipn 26 (enc _ :: _)) pmcells_app; s.
          rewrite (pmcells_cons (_ + 26)).
          rhs rewrite -pmplus_assoc pmplus_swap plus_comm.
          pmplus_eq'.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sjmp].
        - eapply pmread_plus; eauto.
          unfold mt_code, mt_code'.
          rewrite -(firstn_skipn 27 (enc _ :: _)) pmcells_app; s.
          rewrite (pmcells_cons (_ + 27)).
          rhs rewrite -pmplus_assoc pmplus_swap plus_comm.
          pmplus_eq'.
        - econs; tmext_simpl; s.
          eapply pmread_plus; eauto; s; pmplus_eq'.
      }
      { depmcplus0. }

      eapply dance_O; instantiate (1 := fun _ _ _ _ => depmach0); eauto.
    }

    - i; rrdes XTRb RLb; clarify; s in Ec; clarify.
      esplits; i.
      frames_nodep RLa cfas; destruct ea''.
      clear SV HV FIN; destruct_tgt_real RLa.
      
      setcrf mrf'.
      set (s' := (mrf', P, env, s1, switch s2 x (mrf r_sp) (mrf r_ret))).

      assert (EXT': ext (wlift mylw) s' (mrf, P, env, s1, s2)).
      { unfold s'; split.
        - unfold mrf'; repeat split; s; tmext_simpl; eauto.
        - apply switch_ext.
      }

      exists s'; esplits.
      + eauto.
      + rewrite tgt_csplit; eauto.
      + rr; s; eauto.
      + econs 2; econs 1.
        * simpl app_cont.
          inst_pairs; eapply Rok, red_ctx; econs; econs; eauto.
        * replace (subst_exp _ _)
             with (Eapp (Efix 1 1 (Eread (Eloc x))) (Eapp (Efix xf xa e) Eunit))
          ; last first.
          { symmetry; apply subexp_id_closed; rewrite -H in CLO0.
            ss; i; des; eauto using CLO0.
          }
          change (app_cont _ _)
            with (app_cont (SKappR (Vfix 1 1 (Eread (Eloc x))) :: kb)
                           (Eapp (Efix xf xa e) Eunit)).
          inst_pairs; eapply Rok, red_ctx; econs; econs.
      + match goal with |- context[app_cont ?a ?b]=>
          instantiate (2:= (pmemp, Nothing, Just (app_cont a b)))
        end.
        instantiate (1 := (_, Just _, Nothing)). s. noemp.
        rewrite -!pmplus_assoc pmplus_swap in PV1, FIN2.
        rr; esplits.
        * eauto.
        * (lhs rewrite -pmplus_assoc 2!pmplus_assoc); eauto.
        * rewrite -!pmplus_assoc pmplus_swap; eauto.
        * rewrite -!pmplus_assoc pmplus_swap; eauto.
      + rr; s; eauto.
      + rr; esplits; last 2 first.
        * instantiate (2 := {| cpc := Nothing; crf := Just mrf'
                             ; cst := freest0; chp := hp ## refha |}).
          instantiate (1 := mkconf _ Nothing _ _). s.
          s; (rhs rewrite <-(pmplus_comm (hp ## _)), pmplus_swap, (pmplus_swap {_|=>_})). eauto.
        * instantiate (2 := (refhb, Just env, Nothing)).
          instantiate (1 := (_, Nothing, _)).
          s; (rhs rewrite -pmplus_assoc pmplus_swap); eauto.
        * { rr; esplits; first split.
            - rewrite pmcells_sng in CPa0.
              eapply tgt_confpred_push; eauto.
              unfold mrf'; tmext_simpl; eauto.
            - eauto.
            - dup EXT'; rrdes EXT'; unfold s'; clean.
              eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto; inst
              ; (s; i; eapply gk_mon; eauto; split; eauto). 
            - s; noemp; eauto.
            - s; noemp; eauto.
          }
        * { unfold s'.
            rewrite -EXT1.
            rr; esplits; s; eauto; [].
            i; apply in_map_iff in IN1; des; destruct x0.
            des_ifH IN1; clean; clarify.
            - subst lb called0; eauto.
              decomp_pmvalid PV.
              eapply pmread_plus; eauto using pmvalid_plus_cell.
            - apply INS, pmread_plusD in IN2; des.
              { exfalso; apply pmread_cellD in IN2; des; apply n0; eauto. }
              decomp_pmvalid PV.
              eapply pmread_plus; eauto using pmvalid_plus_cell.
              rewrite <-pmplus_comm; eauto.
          }
      + s in H0; clarify.       
        eapply call_app; first last; first 1 last.
        * econs; eapply vclgen_unit; rr; eauto.
        * invc RV0 REL; inv VR; clarify.
          eapply vclos_thunk_mon; eauto using (gk_mon GK0).
        * rewrite H; econs; eauto.
        * replace (mrf r_arg) with (mrf' r_clo)
          ; [|by unfold mrf'; tmext_simpl; auto].
          econs; unfold s'; s; unfold mrf'; tmext_simpl; eauto.
        * econs; i.
          invc RETa RETb.

          destruct s'0 as [[[[]]]]; clean.

          assert (NE: encrypted s4 = false).
          { rrdes EXT2 EXT3.
            clear - ENC0; unfold s' in *; clean.
            destruct (encrypted s4); eauto.
          }

          unfold mrf'; tmext_simpl.

          
          clear PV FIN FIN0 SV HV.

          apply pesim_gknow; i.
          apply pesim_acquire.
          ts_prepare.

          eapply pesim_dance; first (by eauto).
          { dup EXT2; rrdes EXT2 EXT2 EXT2 EXT2; rrdes EXT4 EXT6; clean.
            rewrite EXT4. subst.
            unfold s'; s. rewrite EXT1.
            rewrite plus_comm.
            clear SV HV FIN.

            dance_init.
            { eapply Sok; [s; i; omega|eapply Sopr].
              - eapply pmread_plus; eauto.
                unfold mt_code, mt_code'.
                rewrite -(firstn_skipn 28 (enc _ :: _)) pmcells_app; s.
                rewrite (pmcells_cons (_ + 28)).
                rhs rewrite -pmplus_assoc pmplus_swap plus_comm.
                pmplus_eq'.
              - eauto.
              - eauto.
              - eauto.
            }
            { depmcplus0. }

            dance_init.
            { eapply Sok; [s; i; omega|eapply Sld].
              - eapply pmread_plus; eauto.
                unfold mt_code, mt_code'.
                rewrite -(firstn_skipn 29 (enc _ :: _)) pmcells_app; s.
                rewrite (pmcells_cons (_ + 29)).
                rhs rewrite -pmplus_assoc pmplus_swap plus_comm.
                pmplus_eq'.
              - rrdes EXTP EXTP0; s in ST; rewrite -ST.
                rewrite pmplus_comm -2!pmplus_assoc pmplus_swap.
                rrdes EXTP EXTP EXTP.
                unfold s' in *; clean.
                rewrite -RF. unfold mrf'; tmext_simpl.
                rewrite plus_comm; s; rewrite -minus_n_O.
                econs.  tmext_simpl.  s. eapply pmread_plus; eauto.
                pmplus_eq'.
                move SV0 at bottom. rewrite -ST in SV0.
                rewrite (pmplus_comm {_|->_}) pmplus_swap
                        (pmplus_comm freest) -pmplus_assoc in SV0; eauto.
            }
            { depmcplus0. }

            dance_init.
            { eapply Sok; [s; i; omega|eapply Sld].
              - eapply pmread_plus; eauto.
                unfold mt_code, mt_code'.
                rewrite -(firstn_skipn 30 (enc _ :: _)) pmcells_app; s.
                rewrite (pmcells_cons (_ + 30)).
                rhs rewrite -pmplus_assoc pmplus_swap plus_comm.
                pmplus_eq'.
              - eauto.
            }
            { depmcplus0. }

            dance_init.
            { eapply Sok; [s; i; omega|eapply Sjmp].
              - eapply pmread_plus; eauto.
                unfold mt_code, mt_code'.
                rewrite -(firstn_skipn 31 (enc _ :: _)) pmcells_app; s.
                rewrite (pmcells_cons (_ + 31)).
                rhs rewrite -pmplus_assoc pmplus_swap plus_comm.
                pmplus_eq'.
              - eauto.
            }
            { depmcplus0. }

            eapply dance_O; instantiate (1 := fun _ _ _ => depmach0); eauto.
          }

          { clear RLb.
            i; rrdes XTRb RLb; clarify; s in Ec; clarify.

            dup EXT2; rrdes EXT2 EXT2 EXT2 EXT2; rrdes EXT4 EXT6; clean.
            hexploit INS2.  
            { unfold s'; s; apply in_map_iff.
              econs; inst_pairs; s; split; eauto; des_if; vauto.
            }
            i; des; destruct b'0; vauto.

            esplits; first 1 last.
            - left; econs 2; econs 1.
              + inst_pairs; eapply Rok, red_ctx; econs; econs.
              + simpl app_cont.
                inst_pairs; eapply Rok, red_ctx; econs; econs.
                eapply pmread_plus; [eapply INS1| |]; eauto.
                (rhs rewrite pmplus_comm); pmplus_eq'.
            - inst_pairs; s; noemp.
              rr; esplits; eauto.
            - i; frames_nodep RLa cfas; destruct ea'.
              clear SV HV FIN; destruct_tgt_real RLa.
              setcrf mrf''.
              rrdes EXTP EXTP0; s in ST. rewrite -ST.

              set (s4' := {| laddr := laddr s4; encrypted := encrypted s4;
                             instances := instances s4; retaddrs := retaddrs s2 |}).
              set (s'' := (mrf'', P0, env, s3, s4')).

              assert (E: mrf0 r_sp = mrf r_sp + 1 /\ mrf0 r_env = mrf r_env).
              { rrdes EXTP EXTP EXTP; clean.
                rewrite -RF -RF0; s; unfold mrf'; tmext_simpl; auto.
              }
              des.

              assert (SP'': mrf'' r_sp = mrf r_sp).
              { unfold mrf''; tmext_simpl; rewrite E; omega. }

              assert (EXT'': ext (wlift mylw) s'' (mrf0, P0, env, s3, s4)).
              { unfold s''; split.
                - unfold mrf''; repeat split; s; tmext_simpl; eauto.
                - rr; splits; s; eauto.
              }

              assert (EXTP'': extpub (wlift mylw) s'' (mrf, P, env, s1, s2)).
              { unfold s''; split.
                - unfold mrf''; repeat split; s; tmext_simpl; eauto.
                  rewrite E; omega.
                - rr; splits; s; eauto.
                  eapply myext_trans; [by eauto|].
                  eapply myext_trans; [eapply switch_ext|]; eauto.
              }

              exists s''; splits; eauto.

              
              apply esim_fold; econs; i.
              esplits; eauto; [].
              destruct cfa; rrdes XTRa; s in XTRa; s in XTRa0; clarify.
              destruct cfb; rrdes XTRb; clarify.
              invc CNF; des; subst; noemp.
              rrdes MB; s in Ec; clarify.
              destruct ma; clear SV HV FIN; simpl cplus in MA.
              clear ST. unfold s''.
              subst env0 e0 h3.
              destruct_tgt_real MA.
              eapply esim_ret; eauto.
              + econs; eauto.
              + instantiate (2 := Vnat 1).
                inst_pairs; rr; eauto.
              + econs; eapply vclgen_nat; s; eauto.
                unfold mrf''; tmext_simpl.
                split; s; eauto.
              + econs; first 1 last.
                * split; rr; s; eauto.
                * instantiate (1 := mkconf Nothing (Just _) _ _).
                  s; noemp; rr; esplits; s; eauto.
                * instantiate (1 := (_, Just _, Nothing)).
                  s; noemp; rr; esplits; s; eauto.
                * { rr; esplits; first 1 last; first last.
                    
                    - rr; esplits; first split.
                      + clean; eapply tgt_confpred_pop with (popped := [_]); eauto.
                        unfold mrf''; tmext_simpl; s; eauto.
                        rewrite -!pmplus_assoc in SV; decomp_pmvalid SV.
                        rewrite SP''; rewrite <-pmcells_sng, pmplus_comm; eauto.
                      + s; eauto.
                      + dup EXT''; rrdes EXT''; clean.
                        eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto; inst
                        ; (s; i; eapply gk_mon; eauto; split; eauto). 
                      + s; noemp; eauto.
                      + s; noemp; eauto.
                    - instantiate (1 := (_, Nothing, _)); s; eauto.
                    - instantiate (1 := mkconf _ Nothing _ _); s.
                      rewrite <-!pmplus_assoc, (pmplus_swap freest).
                      rewrite SP'' pmcells_sng; eauto.
                    - rr; esplits; eauto.
                      s; rewrite -EXT1 EXT4 NE; auto.
                  }
          }
    - eauto using wfo_le_n_add_S.
  }

  destruct (encrypted s2) eqn:CRY.

  - (*| Code is still encrypted. *)
    rewrite -pmplus_assoc.
    eapply awkward_encrypt; last by apply wfo_le_n_add_S.
    eapply MAIN.

  - (*| Code has been decrypted. *)
    eapply pesim_dance.
    { instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto.
      + rewrite EXT1 (pmplus_comm refha) (pmplus_swap hp) in HV, FIN; eauto.
        
      + rewrite EXT1 (pmplus_comm refha) (pmplus_swap hp) in HV, FIN; eauto.
        
    }
    { clear SV HV FIN.

      dance_init.
      { eapply Sok; [s; i; omega|eapply Slpc].
        eapply pmread_plus; eauto.
        rewrite plus_comm; s.
        unfold mt_code; rewrite !pmcells_cons; simpl negb; pmplus_eq''.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          rewrite plus_comm; s.
          unfold mt_code; rewrite !pmcells_cons; simpl negb; pmplus_eq''.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }


      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - eapply pmread_plus; eauto.
          rewrite plus_comm; s.
          unfold mt_code; rewrite !pmcells_cons; simpl negb; pmplus_eq''.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sjmp].
        - eapply pmread_plus; eauto.
          rewrite plus_comm; s.
          unfold mt_code; rewrite !pmcells_cons; simpl negb; pmplus_eq''.
        - eauto.
      }
      { depmcplus0. }

      eapply dance_O; instantiate (1 := fun _ _ _ => depmach0); eauto.
    }

    clear RLb.
    i; rrdes XTRb RLb; clarify; s in Ec; clarify.
    esplits; last 2 first.
    + right; splits; eauto.
eapply wfo_le_n_S_rev; last first.
eapply wfo_le_n_S_rev; last first.
eapply wfo_le_n_S_rev; last first.
eapply wfo_le_n_S_rev; last first.
eauto.

      eapply wfo_add_ltc1, wfo_nat_lt_mon, lt_n_Sn.
      eapply wfo_add_ltc1, wfo_nat_lt_mon, lt_n_Sn.
      eapply wfo_add_ltc1, wfo_nat_lt_mon, lt_n_Sn.
      eapply wfo_add_ltc1, wfo_nat_lt_mon, lt_n_Sn.
    + inst_pairs; s; noemp; rr; eauto 10.
    + i; frames_nodep RLa cfas; destruct ea'.
      clear SV HV FIN; destruct_tgt_real RLa.
      esplits; eauto.
      replace (Just mrf) with (Just [mrf r_aux//r_aux]>>[mrf r_i//r_i]>>mrf)
      ; [|by rewrite -2!tmext_self; auto].
      replace (_ + 6 - 6) with (laddr (snd s)); last by set (xx := laddr _); omega.
      replace ([laddr (snd s) + 23 // r_aux]>> [laddr (snd s) // r_j]>> mrf)
         with ([mrf r_clo//r_clo]>>[laddr (snd s) + 23 // r_aux]>>[mrf r_i//r_i]>>[laddr (snd s) // r_j]>> mrf)
      ; last first.
      { rewrite (@tmext_swap _ _ _ r_i); [|by vauto].
        rewrite -tmext_self.
        rewrite (@tmext_swap _ _ _ r_clo); [|by vauto].
        rewrite (@tmext_swap _ _ _ r_clo); [|by vauto].
        rewrite -tmext_self; auto.
      }
      eapply MAIN.
Grab Existential Variables.
  exact wfo_zero. exact wfo_zero. exact wfo_zero. exact wfo_zero.
Opaque mt_code. Opaque mt_code'. Opaque mt_code_tbe.
Qed. (** These Opaque commands are for working around a bug in Coq. *)
Transparent mt_code. Transparent mt_code'. Transparent mt_code_tbe.


Lemma awkward_lsim: forall g (s: state (wlift mylw)) vt
    (LA: ts_db s (ts_fun (laddr (snd s))) vt)
    (NZ: laddr (snd s) <> 0),
  lsim (wlift mylw) g s (FTarrow Tunit (Tarrow (Tarrow Tunit Tunit) Tnat)) vt vs.
Proof.
  i; econs.
  { econs; rr; auto.
    unfold vs; splits; rr; s; i; des; vauto.
  }

  i; invc STK; invc RF VR RK; rrdes CA; clear CB.
  invc APPa APPb; unfold vs in VAL; s in VAL; clarify; s; tmext_simpl.
  destruct s' as [[[[]]]]; clean; clear dependent v'.
  apply EXT in LA; clean.

  apply pesim_acquire.
  rename PC into CLO.             
  ts_prepare.

  hexploit tgt_vquery_fun_uniq; [by eauto|exact CLO|exact LA|]; []; i; subst mpc.

  dup EXT; rrdes EXT EXT1; clean; rewrite -EXT1.

  eapply pesim_stepb.
  i; rrdes XTR RLb; clarify.
  rrdes RL; s in Ec; noemp; clarify.
  hexploit (@pmfinite_alloc _ _ _ ((refhb ## hb0) ## h)); eauto; []; i; des.
  
  esplits; [|right; esplits].
  { econs 2; econs 1.
    - replace (app_cont _ _)
         with (app_cont (SKappR (Vabs 0 (Vabs 2 (Eapp (Vabs 1 (Eapp (Vabs 1 (Eread (Evar 0))) (Eapp (Evar 2) Eunit)))
                   (Ewrite (Evar 0) (Enat 1))))) :: kb) (Enew (Enat 0))); eauto.
      
      inst_pairs; eapply Rok, red_ctx; econs; econs; eauto. 
    - simpl app_cont; inst_pairs; eapply Rok, red_ctx; econs; econs; eauto.
  }
  { rewrite pmplus_assoc in X |- *.
    inst_pairs; s; noemp; rr; esplits; eauto.
    rewrite -pmplus_assoc; eauto.
  }
  tmext_simpl.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN FIN0.

  - rewrite EXT1.

    dance_init.
    { eapply Sok; [by vauto|eapply Slpc].
      eapply pmread_plus; eauto.
      unfold mt_code, mt_code'; rewrite pmcells_cons; pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - eapply pmread_plus; eauto.
        unfold mt_code; rewrite 2!pmcells_cons; pmplus_eq'.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Salc].
      - eapply pmread_plus; eauto.
        unfold mt_code; rewrite 3!pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - eapply pmread_plus; eauto.
        unfold mt_code; rewrite 5!pmcells_cons; pmplus_eq''.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }
    s; fold newloc.           

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; eauto.
        unfold mt_code; rewrite 6!pmcells_cons; pmplus_eq''.
      - rewrite pmcells_cons -pmplus_assoc; prep_store; eauto.
    }
    { s; tmext_simpl; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sjmp].
      - eapply pmread_plus; eauto.
        unfold mt_code; rewrite 8!pmcells_cons; pmplus_eq''.
      - eauto.
    }
    { depmcplus0. }
    s; fold newloc.           

    eapply dance_O; instantiate (1 := fun _ _ cf _ _ _ => depmach1 cf); s; eauto.

  - i; rrdes XTRb RLb; clarify; s in Ec; clarify.
    esplits; last 2 first.
    + right; splits; eauto using wfo_le_n_add. 
    + inst_pairs; s; noemp; rr; eauto 10.
    + i. frames_dep 2 RLa XTRa cfas. frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      simpl lang_tgt.chp in *; setnewloc; setcrf mrf'.
      set (s2' := {| laddr := laddr s2; encrypted := encrypted s2
                   ; instances := (x, false) :: instances s2
                   ; retaddrs := retaddrs s2 |}).
      set (s' := (mrf', extend P (ts_fun (laddr s2 + 6)) newloc, env, s1, s2')).
      assert (EXT': extpub (wlift mylw) s' (mrf, P, env, s1, s2)).
      { unfold s', s2', mrf'; repeat split; s; tmext_simpl; eauto.
        ii; rr; eauto.
      }

      esplits; [by apply extpub_in_ext, EXT'|].

      
      apply esim_fold; econs; i.
      invc CNF. des; subst; noemp; clean.
      rrdes MA MB.
      destruct cfa, cfb.
      invc XTRa XTRb. clarify. s in Ec. clarify. subst env0 h1 e.
      s in H; s in H0; subst; s in HP.
      destruct ma; s in HP; clarify.
      s in ST; s in RF; s in PC; clarify.
      s in FIN0; s in HV0; s in SV0; clarify. 

      esplits; eauto; []; eapply esim_ret; eauto.
      * econs; eauto.
      * instantiate (2 := (Vabs 2 (Eapp (Vabs 1 (Eapp (Vabs 1 (Eread (Eloc x)))
                                                      (Eapp (Evar 2) Eunit)))
                                        (Ewrite (Eloc x) (Enat 1))))).
        inst_pairs; rr; eauto.
      * econs; econs; econs.
        { unfold s'; rr; s; tmext_simpl; unfold extend; eauto. }
        { rr; splits; s; i; des; eauto. }
        eapply gk_lsim; eauto.
        unfold s', s2'; eapply awkward_lsim_aux; s; tmext_simpl; rr; eauto.
      * econs; first 1 last.
        { split; rr; s; eauto. }
        { instantiate (1 := mkconf Nothing _ _ _); s; noemp.
          rr; esplits; s; eauto.
        }
        { instantiate (1 := (_, _, Nothing)); s; noemp.
          rr; esplits; eauto.
        }
        rr; esplits.
        { rr; esplits; eauto; first (rr; esplits).
          - unfold s'; eapply tgt_confpred_extend; s; eauto.
            + tmext_simpl; auto.
            + rewrite EXT1 pmcells_sng; decomp_pmvalid HV; auto.
          - rr; eauto.
          - dup EXT'; rrdes EXT'; unfold s'; clean.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto; inst
            ; (s; i; eapply gk_mon; eauto; split; eauto).
        }
        { rr; esplits; eauto; [].
          insts $ ({x |-> Vnat 0} ## hb0).
          unfold s', s2'; s; i; decomp_pmvalid PV2.
          des; clarify; try subst lb called; eauto.
          + eapply pmread_plus; eauto.
          + eapply pmread_plus; [apply INS| |]; eauto.
            rewrite ->pmplus_comm; eauto.
        }
        { s; rewrite ->EXT1, <-!pmplus_assoc, pmcells_sng; noemp; auto. }
        { s; noemp; rewrite pmplus_swap; auto. }
Grab Existential Variables.
  exact wfo_zero. exact wfo_zero.
Opaque mt_code. Opaque mt_code'. Opaque mt_code_tbe.
Qed. (** These Opaque commands are for working around a bug in Coq. *)
Transparent mt_code. Transparent mt_code'. Transparent mt_code_tbe.


Theorem awkward_msim:
  msim ts_gw nil [(AWK, Tawk)] mt ms.
Proof.
  ii; exists mylw; i.

  destruct ciga, cila, cigb as [[]], cilb as [[]].
  rrdes LMa LMb; clarify.
  destruct (load _ _ _) as [[gh lh] exports] eqn:Eload; des.
  s in GPC; s in GRF; s in GST0; s in GHP; s in GDJ.
  s in LPC; s in LRF; s in LST; s in LHP; s in LDJ.
  clarify; noemp.
  
  destruct suba, subb; vauto; []; clear SUBa SUBb.

  assert (hp_imports = pmemp); [|subst; noemp].
  { split_pmvalid GDJ; apply pmemp_from_read; eauto.
    ss; i; eapply GDOM; eauto.
  }

  
  set (db q v := q = ts_fun (2 + ancha) /\ v = 1 + ancha).
  evar (env: env).
  set (sig := (rf, db, env, refst_emp) : state ts_gw).
  set (sil := {| laddr := S (S ancha); encrypted := true
               ; instances := nil; retaddrs := pmemp |}).
  exists (sig, sil); splits.
  - vauto.
  - apply mylw_stable.
  - i; rr; esplits; s.
    + split; [|by rr; eauto].
      insts $ {| cpc := Nothing; crf := Just rf
               ; cst := Some (fun _ => Some 0); chp := {1 + ancha |-> 2 + ancha} |}.
      rr; esplits; eauto; s.
      * split; i.
        { econs; apply pmread_some; eauto; des_if; auto; omega. }
        { des; apply pmread_some in H; des_ifs; omega. }
      * i; rrdes IN; subst; s; eauto.
    + apply ts_refconfrel_emp; eauto.
    + s; noemp; f_equal; eapply pm_extensionality; eauto.
      * rr; i; vauto. 
      * split; i; [by apply pmread_some in H; vauto|].
        rewrite (pmread_uniq H (GST0 x)); apply pmread_some; auto.
      * s in Eload; clarify; noemp.
        rewrite -H2 !plus_0_r plus_comm; s.
        rewrite plus_comm pmcells_sng; s; eauto.
    + s; noemp; unfold env; eauto.
  - i; rr; esplits; s; noemp; eauto; i; vauto.
    s in Eload; clarify; noemp.
    rewrite !plus_0_r plus_comm; s.
    rewrite plus_comm pmplus_comm -minus_n_O pmcells_sng; s; eauto.
  - unfold sig; s; i; rrdes FH; clarify.
  - i; sng; esplits; [M|M|]; Mdo by rr; s; eauto.
    ii; econs.
    + rr; econs; apply LE; s; rr; split; [|rewrite (plus_comm _ ancha)]; eauto.
    + rr; splits; [by econs|]; s; i; des; vauto.
    + eapply gk_lsim; eauto.
      apply awkward_lsim; rrdes LE LE0; auto.
      * rewrite LE0; apply LE; unfold sig, db; s; split; auto.
        rewrite plus_comm; auto.
      * rewrite LE0; unfold sig, db; s; eauto.
Grab Existential Variables.
  exact 0.
Opaque mt_code. Opaque mt_code'. Opaque mt_code_tbe.
Qed. (** These Opaque commands are for working around a bug in Coq. *)
Transparent mt_code. Transparent mt_code'. Transparent mt_code_tbe.






