(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

(** Well-founded orders. *)

Require Import Arith Relations Max NPeano Wellfounded List.
Require Import ssr vgtac basis.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".


(** * Definitions. *)



Structure wfo :=
{ bwfo_set :> Type
; bwfo_lt : bwfo_set -> bwfo_set -> Prop
; bwfo_wf : well_founded bwfo_lt
}.
Arguments bwfo_lt {_} _ _.

(** **
  Non-trival Non-negative Well-founded Partially-ordered Monoids (NtNnWfPoMonoids)
 *)

Structure awfo :=
{ wfo_set :> Type


; wfo_zero : wfo_set
; wfo_add  : wfo_set -> wfo_set -> wfo_set

; wfo_add_zero_l: forall i, wfo_add wfo_zero i = i
; wfo_add_zero_r: forall i, wfo_add i wfo_zero = i
; wfo_add_assoc: forall i j k, wfo_add i (wfo_add j k) = wfo_add (wfo_add i j) k


; wfo_lt : wfo_set -> wfo_set -> Prop
; wfo_le  := fun i j => rtc wfo_lt i j

; wfo_add_lt1: forall i i' j (LT: wfo_lt i i'), wfo_lt (wfo_add i j) (wfo_add i' j)
; wfo_add_lt2: forall i j j' (LT: wfo_lt j j'), wfo_lt (wfo_add i j) (wfo_add i j')


; wfo_wf : well_founded wfo_lt


; wfo_non_negative: forall i, wfo_le wfo_zero i


; wfo_one  : wfo_set
; wfo_non_trivial : wfo_zero <> wfo_one
}.

Hint Unfold wfo_le.

Arguments wfo_lt {_} _ _.
Arguments wfo_le {_} _ _.
Arguments wfo_zero {_}.
Arguments wfo_one {_}.
Arguments wfo_add {_} _ _.



Lemma symprod_tc1: forall A B lt1 lt2 a a' b
    (LT: tc lt1 a a' : Prop),
  tc (@symprod A B lt1 lt2) (a,b) (a',b).
Proof.
  i; induction LT; eauto using t_step, t_trans, left_sym.
Qed.

Lemma symprod_tc2: forall A B lt1 lt2 a b b'
    (LT: tc lt2 b b' : Prop),
  tc (@symprod A B lt1 lt2) (a,b) (a,b').
Proof.
  i; induction LT; eauto using t_step, t_trans, right_sym.
Qed.

Program Definition awfo_symprod (a b : awfo) : awfo :=
{| wfo_lt   := symprod a b wfo_lt wfo_lt
 ; wfo_zero := (wfo_zero, wfo_zero)
 ; wfo_one  := (wfo_one, wfo_zero)
 ; wfo_add  := fun i j => (wfo_add (fst i) (fst j), wfo_add (snd i) (snd j))
|}.
Next Obligation. s; rewrite !wfo_add_zero_l; auto. Qed.
Next Obligation. s; rewrite !wfo_add_zero_r; auto. Qed.
Next Obligation. s; rewrite !wfo_add_assoc; auto. Qed.
Next Obligation. destruct LT; eauto using left_sym, right_sym, wfo_add_lt1, wfo_add_lt2. Qed.
Next Obligation. destruct LT; eauto using left_sym, right_sym, wfo_add_lt1, wfo_add_lt2. Qed.
Next Obligation. eauto using wf_symprod, wfo_wf. Qed.
Next Obligation. 
  assert (NONw := wfo_non_negative _ w).
  assert (NONw0 := wfo_non_negative _ w0).
  apply rtc_tc in NONw; apply rtc_tc in NONw0.
  des; subst; eauto; apply tc_rtc; eauto using t_trans, symprod_tc1, symprod_tc2.
Qed.
Next Obligation. ii; depdes H; eapply wfo_non_trivial; eauto. Qed.



Lemma lexprod_tc1: forall A B lt1 (lt2: relation B) a a' b b'
    (LT: tc lt1 a a' : Prop),
  tc (@lexprod A (fun _ => B) lt1 (fun _ => lt2)) 
     (existT _ a b) (existT _ a' b').
Proof.
  i; induction LT; eauto using t_step, t_trans, left_lex.
Grab Existential Variables.
eauto.
Qed.

Lemma lexprod_tc2: forall A B lt1 lt2 a b b'
    (LT: tc lt2 b b' : Prop),
  tc (@lexprod A (fun _ => B) lt1 (fun _ => lt2)) 
     (existT _ a b) (existT _ a b').
Proof.
  i; induction LT; eauto using t_step, t_trans, right_lex.
Qed.

Program Definition awfo_lexprod (a b : awfo) : awfo :=
{| wfo_lt   := lexprod a (fun _ => b) wfo_lt (fun _ => wfo_lt) 
 ; wfo_zero := existT _ wfo_zero wfo_zero
 ; wfo_one := existT _ wfo_one wfo_zero
 ; wfo_add  := fun i j => existT _ (wfo_add (projT1 i) (projT1 j)) (wfo_add (projT2 i) (projT2 j))
|}.
Next Obligation. s; rewrite !wfo_add_zero_l; auto. Qed.
Next Obligation. s; rewrite !wfo_add_zero_r; auto. Qed.
Next Obligation. s; rewrite !wfo_add_assoc; auto. Qed.
Next Obligation. destruct LT; eauto using left_lex, right_lex, wfo_add_lt1, wfo_add_lt2. Qed.
Next Obligation. destruct LT; eauto using left_lex, right_lex, wfo_add_lt1, wfo_add_lt2. Qed.
Next Obligation. eauto using wf_lexprod, wfo_wf. Qed.
Next Obligation. 
  assert (NONw := wfo_non_negative _ i).
  assert (NONw0 := wfo_non_negative _ X).
  apply rtc_tc in NONw; apply rtc_tc in NONw0.
  des; subst; eauto; apply tc_rtc; eauto using t_trans, lexprod_tc1, lexprod_tc2.
Qed.
Next Obligation. ii; depdes H; eapply wfo_non_trivial; eauto. Qed.



Definition gwf_set : Type := list (sigT bwfo_set).

Definition gwfel (o: wfo) (x: o) :=
  existT _ o x.

Definition gwf_one : gwf_set := 
  [ gwfel (Build_wfo lt_wf) 0 ].

Inductive gwf_lt : gwf_set -> gwf_set -> Prop := 
| gwf_lt_0 el i:
    gwf_lt i (el :: i)

| gwf_lt_1 o x y i
      (LT: bwfo_lt x y):
    gwf_lt (gwfel o x :: i) (gwfel o y :: i)

| gwf_lt_S i j el
      (LE: gwf_lt i j):
    gwf_lt (el :: i) (el :: j)
.
Hint Constructors gwf_lt.

Lemma gwf_lt_succ_inv i j o x
    (LT: gwf_lt i (gwfel o x :: j)):
  i = j \/
  (exists k, i = gwfel o x :: k /\ gwf_lt k j) \/
  (exists y, i = gwfel o y :: j /\ bwfo_lt y x).
Proof.
  by ginduction i; i; depdes LT; eauto 10.
Qed.

Program Definition gwfo : awfo :=
{| wfo_set  := gwf_set
 ; wfo_lt   := gwf_lt
 ; wfo_zero := nil
 ; wfo_one  := gwf_one
 ; wfo_add x y := x ++ y
|}.
Next Obligation. apply app_nil_r. Qed.
Next Obligation. apply app_assoc. Qed.
Next Obligation. induction LT; s; eauto. Qed.
Next Obligation. ginduction i; s; eauto. Qed.
Next Obligation.
  intro i; induction i.
  - econs; i; depdes H.
  - destruct a; assert (ACCb := bwfo_wf x b).
    ginduction ACCb; i.
    ginduction IHi; i.
    econs; i; apply gwf_lt_succ_inv in H3.
    des; subst; eauto using Acc_intro.
Qed.
Next Obligation. induction i; eauto using rt_trans. Qed.



Definition wfo_embed {o: wfo} (a: o) : gwfo :=
  [ gwfel o a ].

Lemma wfo_embed_preserve (o: wfo) (a b: o)
    (LT: tc bwfo_lt a b):
  tc wfo_lt (wfo_embed a) (wfo_embed b).
Proof.
  induction LT.
  - by apply t_step; econs.
  - eauto using t_trans.
Qed.

Lemma wfo_embed_reflect (o: wfo) (a b: o)
    (LT: tc wfo_lt (wfo_embed a) (wfo_embed b)):
  tc bwfo_lt a b.
Proof.
  revert a b LT.
  assert (INV: forall i j (x: o) 
                 (LT: clos_trans_1n _ wfo_lt i j)
                 (EQ: j = wfo_embed x), 
               i = nil \/ 
               exists y : o, i = wfo_embed y).
  {
    unfold wfo_embed; i; ginduction LT; i; subst.
    - destruct x; inv H; eauto; inv LE.
    - specialize (IHLT _ _ eq_refl); des; subst.
      + inv H.
      + destruct x; inv H; eauto; inv LE.
  }

  i; remember (wfo_embed a) as i.
  remember (wfo_embed b) as j.
  apply clos_trans_t1n in LT.
  ginduction LT; i; subst.
  - inv H; depdes H1; eauto; inv LE.
  - eapply INV in LT; eauto.
    des; subst; [by inv H|].
    eapply t_trans; eauto.
    eapply t_step. 
    inv H; depdes H1; eauto; inv LE.
Qed.



Lemma wfo_add_ltc1: forall (o: awfo) (i i' j: o)
    (LT: tc wfo_lt i i'),
  tc wfo_lt (wfo_add i j) (wfo_add i' j).
Proof. 
  i; induction LT; eauto using t_step, t_trans, wfo_add_lt1.
Qed.

Lemma wfo_add_ltc2: forall (o: awfo) (i j j': o)
    (LT: tc wfo_lt j j'),
  tc wfo_lt (wfo_add i j) (wfo_add i j').
Proof. 
  i; induction LT; eauto using t_step, t_trans, wfo_add_lt2.
Qed.

Lemma wfo_add_monotone: forall (a: awfo) (i i' j j': a)
    (LEi: wfo_le i i')
    (LEj: wfo_le j j'),
  wfo_le (wfo_add i j) (wfo_add i' j').
Proof.
  intro a.
  assert (ADDLE:
    forall (x j j' : a) (LE: wfo_le j j'),
      rtc wfo_lt (wfo_add x j) (wfo_add x j')).
  { i; induction LE; eauto using rt_step, rt_trans, wfo_add_lt2. }

  i; ginduction LEi; i; eauto.
  - i; eapply rt_trans, rt_step, wfo_add_lt1; eauto.
  - i; eapply rt_trans; [eapply IHLEi1|eapply IHLEi2]; eauto.
Qed.

Lemma wfo_add_le1: forall a i j, @wfo_le a i (wfo_add i j).
Proof.
  i; rewrite <-(wfo_add_zero_r _ i) at 1.
  eauto using wfo_add_monotone, wfo_non_negative.
Qed.

Lemma wfo_add_le2: forall a i j, @wfo_le a j (wfo_add i j).
Proof.
  i; rewrite <-(wfo_add_zero_l _ j) at 1.
  eauto using wfo_add_monotone, wfo_non_negative.
Qed.

Lemma wfo_le_eq_or_lt: forall (a: awfo) (i j: a)
    (LE: wfo_le i j),
  i = j \/ tc wfo_lt i j.
Proof.
  i; apply clos_rt_rt1n in LE.
  ginduction LE; des; subst; eauto using t_trans.
Qed.

Lemma wfo_zero_one: forall a,
  tc wfo_lt (@wfo_zero a) wfo_one.
Proof.
  i; assert (LE:= wfo_non_negative a wfo_one).
  apply wfo_le_eq_or_lt in LE; des; eauto.
  exfalso; eapply wfo_non_trivial; eauto.
Qed.

Lemma wfo_add_one: forall (a: awfo) (i: a), tc wfo_lt i (wfo_add wfo_one i).
Proof.
  i; rewrite <-(wfo_add_zero_l a i) at 1.
  eauto using wfo_add_ltc1, wfo_zero_one.
Qed.

(** [wfo_le_n n o o'] means that [o'] can be decreased at least [n] times
    before reaching [o].
*)
Inductive wfo_le_n {o: awfo} : nat -> o -> o -> Prop :=
| wfo_le_n_Z i j (LE: wfo_le i j)
  : wfo_le_n O i j
| wfo_le_n_S n i j k (GLE: wfo_le_n n i j) (LT: tc wfo_lt j k)
  : wfo_le_n (S n) i k
.
Hint Constructors wfo_le_n.

Lemma wfo_le_n_lt: forall o n (i i': wfo_set o)
    (LE: wfo_le_n (S n) i i'),
  tc wfo_lt i i'.
Proof.
  induction n; i; invc LE; [by invc GLE; s in LE; eauto using rtc_tc_tc|].
  econs 2; last first; eauto using IHn.
Qed.


Lemma wfo_le_n_S': forall o n (i i': wfo_set o)
    (LE: wfo_le_n (S n) i i'),
  exists j : wfo_set o,
    <<LT: tc wfo_lt i j>> /\
    <<LE: wfo_le_n n j i'>>.
Proof.
  induction n; i; invc LE.
  - inv GLE; eauto 10 using rtc_tc_tc.
  - hexploit IHn; eauto; i; des; eauto.
Qed.

Lemma wfo_le_n_S'': forall o n (i i': wfo_set o)
    (LE: wfo_le_n (S n) i i'),
   wfo_le_n n i i'.
Proof.
  i; invc LE GLE; [econs 1|]; eauto using rt_trans, tc_rtc, t_trans.
Qed.

Lemma wfo_le_n_wfo_le: forall o n (i i' i'': wfo_set o)
   (GLE: wfo_le_n n i i')
   (LE: wfo_le i' i''),
 wfo_le_n n i i''.
Proof.
  i; depdes GLE; [econs 1|econs 2].
  - econs 3; eauto.
  - eapply rtc_tc in LE; des; subst; eauto.
  - eauto using tc_rtc_tc.
Qed.

Lemma wfo_le_n_anti: forall o m n (i i': wfo_set o)
    (GLE: wfo_le_n m i i')
    (LE: n <= m),
  wfo_le_n n i i'.
Proof.
  i; ginduction LE; [by auto|].
  i; apply IHLE; auto using wfo_le_n_S''.
Qed.


Lemma wfo_le_n_S_rev: forall o (i j: wfo_set o) n k
    (LT: tc wfo_lt i j)
    (GLE: wfo_le_n n j k),
  wfo_le_n (S n) i k.
Proof.
  i; ginduction GLE; i.
  - econs; eauto using tc_rtc_tc.
  - eauto.
Qed.

Lemma wfo_le_n_plus: forall o n na nb (i i': wfo_set o)
    (GLE: wfo_le_n n i i')
    (E: n = na + nb),
  exists ii,
    <<GLEa: wfo_le_n na i ii>> /\
    <<GLEb: wfo_le_n nb ii i'>>.
Proof.
  i; ginduction na; s; i; subst.
  - exists i; eauto.
  - apply wfo_le_n_S' in GLE; des.
    eapply IHna in LE; des; eauto.
    exists ii; eauto; split; eauto using wfo_le_n_S_rev.
Qed.



Fixpoint wfo_nat {o : awfo} (n: nat) : wfo_set o :=
  match n with
  | 0 => wfo_one
  | S n' => wfo_add wfo_one (wfo_nat n')
  end.

Lemma wfo_le_n_add_S: forall o n (i: wfo_set o),
  wfo_le_n (S n) i (wfo_add (wfo_nat n) i).
Proof.
  induction n; eauto using rt_refl, wfo_add_ltc1, wfo_add_one.
Qed.

Lemma wfo_le_n_add: forall o n (i: wfo_set o),
  wfo_le_n n i (wfo_add (wfo_nat n) i).
Proof.
  i; hexploit (wfo_le_n_add_S o n i); eauto; i.
  invc X GLE; [econs 1|econs 2]; eauto using rt_trans, tc_rtc, t_trans.
Qed.

Lemma wfo_nat_le_mon: forall o n m
    (LE: n <= m),
  @wfo_le o (wfo_nat n) (wfo_nat m).
Proof.
  i; induction LE; eauto.
  eapply rt_trans; eauto. 
  apply wfo_add_le2.
Qed.

Lemma wfo_nat_lt_mon: forall o n m
    (LE: n < m),
  tc (@wfo_lt o) (wfo_nat n) (wfo_nat m).
Proof.
  i; induction LE; eauto using t_trans, t_step, wfo_add_one.
Qed.

