(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program Relations Omega.
Require Export  paco9  vgtac basis countable gwfo.
Require Import lang_common  lang_spec model_common.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".





(** * Relation Types *)



Definition vrelf {La Lb} :=  La.(val) -> Lb.(val) -> Prop.
Definition vrel {La Lb} :=  La.(val) -> Lb.(val) -> Prop.
Definition krel {La Lb} :=  La.(cont)->Lb.(cont) -> Prop.
Definition crel {La Lb} := La.(conf) -> Lb.(conf) -> Prop.

Notation "p <r= q" :=
  (forall  va vb (PR: p  va vb : Prop), 
          q  va vb : Prop)
  (at level 50, no associativity).

Notation "p <g= q" :=
  (forall s, (p s) <r= (q s))
  (at level 50, no associativity).

Notation "p <ks= q" :=
  (forall  k1 k2 (PR: p  k1 k2 : Prop), 
          q  k1 k2 : Prop)
  (at level 50, no associativity).

Notation "p \r/ q" :=
  (fun  va vb => 
     p  va vb \/ q  va vb)
  (at level 49, no associativity).

Notation "p \g/ q" :=
  (fun s => (p s) \r/ (q s))
  (at level 49, no associativity).

Notation "p /r\ q" :=
  (fun  va vb => 
     p  va vb /\ q  va vb)
  (at level 49, no associativity).

Notation "p /g\ q" :=
  (fun s => (p s) /r\ (q s))
  (at level 49, no associativity).

Definition botr {La Lb} : @vrelf La Lb :=
  fun  _ _ => False.


(** * World *)


Structure method_query La Lb (sts: transys) :=
{ vquerya: sts.(state) -> @vquery La -> La.(val) -> Prop
; vqueryb: sts.(state) -> @vquery Lb -> Lb.(val) -> Prop
; cquerya: sts.(state) -> @cquery La -> La.(conf) -> Prop
; cqueryb: sts.(state) -> @cquery Lb -> Lb.(conf) -> Prop
; rquery: sts.(state) -> @vrel La Lb

; vquerya_mon: forall s q v (WV: vquerya s q v) s' (R: sts.(ext) s' s),
                 vquerya s' q v
; vqueryb_mon: forall s q v (WV: vqueryb s q v) s' (R: sts.(ext) s' s),
                 vqueryb s' q v
; cquerya_core: forall s q c (CQ: cquerya s q c), core c
; cqueryb_core: forall s q c (CQ: cqueryb s q c), core c
; rquery_mon: forall s  va vb (WV: rquery s  va vb)
                   s' (R: sts.(ext) s' s),
              rquery s'  va vb

}.



Structure method_conf La Lb (sts: transys) :=
{ confrel : (sts.(state) -> @vrelf La Lb) -> sts.(state) -> @crel La Lb

; confrel_mon : forall g s ha hb (PR: confrel g s ha hb) 
                g' (LE: g <g= g'),
              confrel g' s ha hb
}.

Structure world La Lb :=
{ wwfo  :> awfo
; wsts  :> transys
; wquery :> method_query La Lb wsts

; wconf :> method_conf La Lb wsts





}.

Structure world_glob La Lb :=
{ gwsts  :> transys
; gwquery :> method_query La Lb gwsts
; gwconf :> method_conf La Lb gwsts
}.


Structure world_loca La Lb (gsts: world_glob La Lb) :=
{ lwwfo  :> awfo
; lwsts  :> transys

; lwconf :> method_conf La Lb (tsprod gsts lwsts)






; lwconf_extra: forall g s ha hb (CR: lwconf.(confrel) g s ha hb),
                  extra ha /\ extra hb
}.


(** * Operations on Worlds *)


Program Definition mqlift La Lb tsa (mg: method_query La Lb tsa) tsb
 : method_query La Lb (tsprod tsa tsb) :=
{| vquerya s := mg.(vquerya) (fst s)
 ; vqueryb s := mg.(vqueryb) (fst s)
 ; cquerya s := mg.(cquerya) (fst s)
 ; cqueryb s := mg.(cqueryb) (fst s)
 ; rquery  s := mg.(rquery) (fst s)
|}.
Next Obligation. ss; eauto using vquerya_mon. Qed.
Next Obligation. ss; eauto using vqueryb_mon. Qed.
Next Obligation. ss; eauto using cquerya_core. Qed.
Next Obligation. ss; eauto using cqueryb_core. Qed.
Next Obligation. ss; eauto using rquery_mon. Qed.


Inductive gmon {La Lb} sts (g: sts.(state) -> vrelf) s : @vrelf La Lb :=
| gmon_: forall s'  va vb
    (REL: g s'  va vb)
    (EXT: ext _ s s'),
  gmon sts g s  va vb
.
Hint Constructors gmon.

Lemma gmon_mon: forall La Lb sts g s  va vb
    (REL: @gmon La Lb sts g s  va vb)
    g' (LE: g <g= g')
    s' (EXT: ext _ s' s),
  gmon sts g' s'  va vb.
Proof.
  i; destruct REL; eauto using ext_trans.
Qed.


Definition prod_gka La Lb gw stsa stsb
           (g: state (tsprod gw (tsprod stsa stsb)) -> @vrelf La Lb)
           (slb: state stsb)
  : state (tsprod gw stsa) -> vrelf :=
  fun sa => gmon _ (fun slb' => g (fst sa, (snd sa, slb'))) slb.
Hint Unfold prod_gka.

Definition prod_gkb La Lb gw stsa stsb
           (g: state (tsprod gw (tsprod stsa stsb)) -> @vrelf La Lb)
           (sla: state stsa)
  : state (tsprod gw stsb) -> vrelf :=
  fun sb => gmon _ (fun sla' => g (fst sb, (sla', snd sb))) sla.
Hint Unfold prod_gkb.




Definition crelplus La Lb (rx ry: @crel La Lb) : @crel La Lb := fun ha hb =>
  exists hax hay hbx hby,
  <<WHX: rx hax hbx>> /\ <<WHY: ry hay hby>> /\
  <<HPA: hax # hay = ha>> /\ <<HPB: hbx # hby = hb>>.

Program Definition conflift La Lb gw (lw: world_loca gw) 
  : method_conf La Lb (tsprod gw lw) :=
{| confrel g s := crelplus (gw.(confrel) (fun sg => g (sg, snd s)) (fst s))
                           (lw.(confrel) g s)
|}.
Next Obligation.
  ss; rdes PR; r; esplits; eauto using confrel_mon.
Qed.

Program Definition confprod La Lb gw (lwa lwb: world_loca gw) 
  : method_conf La Lb (tsprod gw (tsprod lwa lwb)) :=
{| confrel g s := crelplus (lwa.(confrel) (prod_gka g (snd (snd s))) (fst s, fst (snd s)))
                           (lwb.(confrel) (prod_gkb g (fst (snd s))) (fst s, snd (snd s)))
|}.
Next Obligation.
  ss; rdes PR; r; esplits; eauto; eapply confrel_mon; eauto
  ; inst; ss; eauto using gmon_mon.
Qed.

Definition wlift La Lb gw (lw: world_loca gw) : world La Lb :=
{| wwfo := lwwfo lw
 ; wsts  := tsprod gw lw
 ; wquery := mqlift gw.(gwquery) lw

 ; wconf  := conflift lw
|}.

Program Definition wprod La Lb (gw: world_glob La Lb) (lwa lwb: world_loca gw) 
 : world_loca gw :=
{| lwwfo := awfo_symprod (lwwfo lwa) (lwwfo lwb)
 ; lwsts := tsprod lwa lwb

 ; lwconf := confprod lwa lwb
|}.
Next Obligation.
  rrdes CR; subst; eapply lwconf_extra in WHX; eapply lwconf_extra in WHY.
  des; eauto using extra_plus.
Qed.


(** * Value Closure *)


Inductive vclos_thunk La Lb sts (mg: method_query La Lb sts) s (g: vrelf) 
 : @vrelf La Lb :=
| vclf_fix: forall 
    va (CA: mg.(vquerya) s valid_fun va)
    vb (CB: mg.(vqueryb) s valid_fun vb)
    (VR: g  va vb),
  vclos_thunk mg s g  va vb
| vclf_gen: forall 
    va (CA: mg.(vquerya) s valid_gen va)
    vb (CB: mg.(vqueryb) s valid_gen vb)
    (VR: g  va vb),
  vclos_thunk mg s g  va vb
.                    
Hint Constructors vclos_thunk.

Inductive vclos_ La Lb sts (mg: method_query La Lb sts) s (r: vrelf) (vr: vrel)
 : @vrel La Lb :=
| vclgen_unit: forall 
    va (CA: mg.(vquerya) s valid_unit va) 
    vb (CB: mg.(vqueryb) s valid_unit vb),
  vclos_ mg s r vr  va vb
| vclgen_nat: forall n
    va (CA: mg.(vquerya) s (valid_nat n) va) 
    vb (CB: mg.(vqueryb) s (valid_nat n) vb),
  vclos_ mg s r vr  va vb
| vclgen_pair: forall 
    vax vay va (CA: mg.(vquerya) s (valid_pair vax vay) va) 
    vbx vby vb (CB: mg.(vqueryb) s (valid_pair vbx vby) vb)
    (VRX: vr  vax vbx) 
    (VRY: vr  vay vby),
  vclos_ mg s r vr  va vb
| vclgen_inl: forall 
    vax va (CA: mg.(vquerya) s (valid_inl vax) va)
    vbx vb (CB: mg.(vqueryb) s (valid_inl vbx) vb)
    (VRX: vr  vax vbx),
  vclos_ mg s r vr  va vb
| vclgen_inr: forall 
    vay va (CA: mg.(vquerya) s (valid_inr vay) va)
    vby vb (CB: mg.(vqueryb) s (valid_inr vby) vb)
    (VRY: vr  vay vby),
  vclos_ mg s r vr  va vb
| vclgen_roll: forall 
    vam va (CA: mg.(vquerya) s (valid_roll vam) va)
    vbm vb (CB: mg.(vqueryb) s (valid_roll vbm) vb)
    (VR: vr  vam vbm),
  vclos_ mg s r vr  va vb
| vclgen_pack: forall 
    
    vam va (CA: mg.(vquerya) s (valid_pack vam) va)
    vbm vb (CB: mg.(vqueryb) s (valid_pack vbm) vb)
    (VR: vr  vam vbm),
  vclos_ mg s r vr  va vb
| vclgen_ref: forall 
    va
    vb
    (VR: mg.(rquery) s  va vb),
  vclos_ mg s r vr  va vb
| vclgen_name : forall 
    va 
    vb (CB: mg.(vqueryb) s valid_name vb)
    (VR: r  va vb),
  vclos_ mg s r vr  va vb
| vclgen_thunk: forall 
    va
    vb
    (VR: vclos_thunk mg s r  va vb),
  vclos_ mg s r vr  va vb
.
Local Hint Constructors vclos_. 

Inductive vclos La Lb sts (mg: method_query La Lb sts) s r  va vb 
 : Prop :=
| vclos_intro 
    (REL: vclos_ mg s r (vclos mg s r)  va vb).
Hint Constructors vclos.


(** * Program Simulation *)


Definition esimtyp {La Lb} (w: world La Lb) := 
  forall (i: wwfo w) (kia: La.(cont)) (kib: Lb.(cont))
    (g: w.(state) -> @vrelf La Lb) (si s: w.(state)) (sync: bool)
     (ea: La.(conf)) (eb: Lb.(conf)), Prop.

Definition ksimtyp {La Lb} (w: world La Lb) := 
  forall (i: wwfo w) (kia: La.(cont)) (kib: Lb.(cont))
    (g: w.(state) -> @vrelf La Lb) (si sp s: w.(state)) 
     (ka: La.(cont)) (kb: Lb.(cont)), Prop.

Notation "p <e= q" :=
  (forall i kia kib g si s sync  ea eb 
     (PR: p i kia kib g si s sync  ea eb : Prop), 
   q i kia kib g si s sync  ea eb : Prop)
  (at level 50, no associativity).

Notation "p \e/ q" :=
  (fun i kia kib g si s sync  ea eb => 
     p i kia kib g si s sync  ea eb \/
     q i kia kib g si s sync  ea eb)
  (at level 50, no associativity).

Notation "p <k= q" :=
  (forall i kia kib g si sp s  ka kb 
     (PR: p i kia kib g si sp s  ka kb : Prop), 
   q i kia kib g si sp s  ka kb : Prop)
  (at level 50, no associativity).

Notation "p \k/ q" :=
  (fun i kia kib g si sp s  ka kb => 
     p i kia kib g si sp s  ka kb \/
     q i kia kib g si sp s  ka kb)
  (at level 50, no associativity).

Definition bote_ {La Lb w} : @esimtyp La Lb w :=
  fun _ _ _ _ _ _ _  _ _ => False.

Notation bote := (pacoid bote_).

Inductive configure {La Lb} {sts} (m: @method_conf La Lb sts) g s (sync: bool)
  ea eb csa csb cfa cfb ma mb : Prop :=
_configure
  (RH: if sync then m.(confrel) g s csa csb else csa = cemp /\ csb = cemp)
  (CO: if sync then core ea /\ core eb else True)
  (MA: La.(real) (ea # csa # cfa) ma)
  (MB: Lb.(real) (eb # csb # cfb) mb)
.
Hint Constructors configure.

Inductive call {La Lb} (w: world La Lb) s rf rv rk  ea eb
 : Prop :=
| call_app
     fa fb va vb ka kb
    (APPa: w.(cquerya) s (app fa va ka) ea)
    (APPb: w.(cqueryb) s (app fb vb kb) eb)
    (RF: vclos_thunk w s rf  fa fb)
    (RV: vclos w s rv  va vb)
    (RK: rk  ka kb : Prop)

| call_inst
     ga gb ka kb
    (INSTa: w.(cquerya) s (inst ga ka) ea)
    (INSTb: w.(cqueryb) s (inst gb kb) eb)
    (RF: vclos_thunk w s rf  ga gb)
    
    (RK: rk  ka kb)
.
Hint Constructors call.

Inductive vsingle {La Lb}  va vb : @vrelf La Lb :=
  _vsingle: vsingle  va vb  va vb.
Hint Constructors vsingle.

Inductive ksingle {La Lb} ka kb : @krel La Lb :=
  _ksingle:  ksingle ka kb  ka kb.
Hint Constructors ksingle.

Inductive good_thunk {La Lb} (w: world La Lb) s : @vrelf La Lb :=
| gt_fun: forall  fa fb 
    (WFa: w.(vquerya) s good_fun fa) 
    (WFb: w.(vqueryb) s good_fun fb),
  good_thunk w s  fa fb
| gt_gen: forall  fa fb 
    (WFa: w.(vquerya) s good_gen fa)
    (WFb: w.(vqueryb) s good_gen fb),
  good_thunk w s  fa fb
. 
Hint Constructors good_thunk.

Inductive lsim_ {La Lb w} (es: @esimtyp La Lb w)
                (g: w.(state) -> vrelf) (s: w.(state)) : @vrelf La Lb :=
_lsim  fa fb i
    (WF: good_thunk w s  fa fb)
    (CS: forall kia kib
           g' (LE: g <g= g')
           s' (EXT: w.(ext) s' s)
            ea eb (STK: call w s' (vsingle  fa fb) (g' s')
                           (ksingle kia kib)  ea eb),
         es i kia kib g' s' s' true  ea eb)
  : lsim_ es g s  fa fb
.
Hint Constructors lsim_.

Inductive ksim_ {La Lb w} (es: @esimtyp La Lb w) : ksimtyp w := 
_ksim i kia kib g si sp s  ka kb
  (KR: forall g' (LE: g <g= g')
           s' (EXTP: w.(extpub) s' sp) (EXT: w.(ext) s' s)
           va vb (RV: vclos w s' (g' s')  va vb)
           ea (RETa: w.(cquerya) s' (ret va ka) ea)
           eb (RETb: w.(cqueryb) s' (ret vb kb) eb),
         es i kia kib g' si s' true  ea eb : Prop)
  : ksim_ es i kia kib g si sp s  ka kb.
Hint Constructors ksim_.

Structure gknow_ {La Lb} w (es: @esimtyp La Lb w)
  (g: w.(state) -> @vrelf La Lb) : Prop
:=
{ gk_mon : forall s  va vb (R: g s  va vb)
             s' (R: w.(ext) s' s),
           g s'  va vb
; gk_lsim : lsim_ es g <g= g

}.
Hint Constructors gknow_.
Implicit Arguments gknow_ [La Lb].


Inductive esim_progress {La Lb w} (es: @esimtyp La Lb w) (i: wwfo w) kia kib
  g si s sync evt mb mb'  ea eb: Prop :=
_esim_progress (i': wwfo _)
    (ST: <<STEP: Lb.(step) evt mb mb'>> \/
         (<<ET: evt = etau>> /\ <<ST: mb = mb'>> /\ <<LT: tc wfo_lt i' i>>))
    (ES: es i' kia kib g si s sync  ea eb : Prop)
    (SYN: sync = false)
.
Hint Constructors esim_progress.

Inductive esim_call {La Lb w} (es: @esimtyp La Lb w) kia kib
  g si s sync evt mb mb'  ea eb : Prop :=
_esim_call i'
    (ST: Lb.(step) evt mb mb')
    (SU: call w s (g s) (g s) (ksim_ es i' kia kib g si s s)  ea eb)
    (SYN: sync = true)
.
Hint Constructors esim_call.

Inductive esim_main {La Lb} {w: world La Lb} esimstep
  kia kib g si s  cfa ma cfb mb : Prop :=
| esim_err
    mb' (STP: rtc (Lb.(step) etau) mb mb')
    (ER: Lb.(error) mb')
| esim_step
    (NT: ~ La.(halted) ma)
    (EST: forall evt ma' (ST: La.(step) evt ma ma'),
          exists ea' csa' eb' csb' mb' mbt s' sync,
          <<STP: rtc (Lb.(step) etau) mb mb'>> /\
          <<ES: esimstep s' sync evt mb' mbt  ea' eb'>> /\
          <<CNF: configure w g s' sync ea' eb' csa' csb' cfa cfb ma' mbt>> /\
          <<EXT: w.(ext) s' s>>)
| esim_ret s'
    mb' (STP: rtc (Lb.(step) etau) mb mb')
    va ea' (RETa: w.(cquerya) s' (ret va kia) ea')
    vb eb' (RETb: w.(cqueryb) s' (ret vb kib) eb')
    (RV: vclos w s' (g s')  va vb)
    csa' csb' (CNF: configure w g s' true ea' eb' csa' csb' cfa cfb ma mb')
    (EXT: w.(ext) s' s) (EXTP: w.(extpub) s' si)
.
Hint Constructors esim_main.

Inductive esim_ {La Lb w} (es: esimtyp w) i kia kib g si s sync 
                (ea: La.(conf)) (eb: Lb.(conf)) : Prop :=
_esim_
  (ESIM: forall
      (GK: gknow_ w es g)
      cfa cfb (XTRa: extra cfa) (XTRb: extra cfb)
      csa csb ma mb (CNF: configure w g s sync ea eb csa csb cfa cfb ma mb),
    esim_main
      (esim_progress es i kia kib g si \7/ esim_call es kia kib g si)
      kia kib g si s  cfa ma cfb mb)
.
Hint Constructors esim_.

(** _Respectful_ closure for a given world *)
Inductive resclo {La Lb w} (clo: @esimtyp La Lb w -> esimtyp w) : Prop :=
_resclo 
  (MON:  monotone9  clo)
  (RESPECT: forall l r (LE: l <e= r) (SIM: l <e= esim_ r),
            clo l <e= esim_ (clo r)).
Hint Constructors resclo.

(** The greatest respectful closure for a given world *)
Inductive gres {La Lb w} (es: @esimtyp La Lb w) 
  i kia kib g si s sync  ea eb : Prop :=
_gres clo
  (RES: resclo clo) 
  (CLO: clo es i kia kib g si s sync  ea eb).
Hint Constructors gres.

Definition pesim {La Lb w} : esimtyp w -> @esimtyp La Lb w := 
   paco9  (esim_ <*> @gres _ _ w).

Definition esim {La Lb} w := @pesim La Lb w bote.
Definition lsim {La Lb} (w: world La Lb) := lsim_ (esim w).
Definition ksim {La Lb} (w: world La Lb) := ksim_ (esim w).
Definition gknow {La Lb} (w: world La Lb) := gknow_ w (esim w).


(** * Module Simulation *)


Definition realizable_global_state La Lb cla clb (gw: world_glob La Lb) g sp :=
  exists ea eb csa csb cfa cfb ma mb,
    configure gw g sp true ea eb csa csb (cla # cfa) (clb # cfb) ma mb.

(** Stability of a local world *)
Definition stable La Lb (gw: world_glob La Lb) (lw: world_loca gw) :=
  forall 
    g (GK: gknow (wlift lw) g)
    sp s cla clb (CR: lw.(confrel) g (sp, s) cla clb)
    sp' (EXT: gw.(ext) sp' sp) 
    (COMPAT: realizable_global_state cla clb gw (fun sg => g (sg, s)) sp'),
  exists s',
    <<EXTP: extpub _ s' s>> /\
    <<CR': lw.(confrel) g (sp', s') cla clb>>.

Definition realizable_local_state La Lb {gw: world_glob La Lb} (lw: world_loca gw) g s :=
  exists ca cb, lw.(confrel) g s ca cb. 

Definition tlsim La Lb (gw: world_glob La Lb) (lw: world_loca gw) ienv suba subb si va vb  :=
  forall s (LE: (wlift lw).(ext) s si)
         g (GK: gknow (wlift lw) g)
         (RLS: realizable_local_state _ g s)
         (RS: forall (f: label)  va' vb'
                     (INienv:  In f ienv )
                     (INsuba: lookup f suba = Some va')
                     (INsubb: lookup f subb = Some vb'),
                vclos_thunk (wlift lw) s (g s)  va' vb'),
    vclos_thunk (wlift lw) s  (g s)  va vb.

(** ** Module simulation *)
Definition msim La Lb (gw: world_glob La Lb) ienv oenv moda modb : Prop :=
  
  exists (lw: world_loca gw),
  forall ancha suba ciga cila (LMa: La.(cload) moda ancha suba ciga cila)
         anchb subb cigb cilb (LMb: Lb.(cload) modb anchb subb cigb cilb)
         (SUBa: map fst suba =  ienv)
         (SUBb: map fst subb =  ienv),
  exists si,
      
      <<STA: stable lw>> /\
      <<INIg: forall g (GK: gknow (wlift lw) g), gw.(confrel) (fun sg => g (sg, snd si)) (fst si) ciga cigb>> /\
      <<INIl: forall g (GK: gknow (wlift lw) g), lw.(confrel) g si cila cilb>> /\
      <<REF: forall  la lb, ~ gw.(rquery) (fst si)  la lb>> /\
      <<SIM: forall f  (IN:  In f oenv ),
               exists va vb,
                 <<Ea: vload _ moda ancha suba f va>> /\
                 <<Eb: vload _ modb anchb subb f vb>> /\
                 <<SIM: tlsim ienv suba subb si va vb >> >>.


(** Induction Principle and Monotonicity *)


Section Simple_Properties_of_Model.

Variables La Lb : LangSpec.
Variable w : world La Lb.

Lemma configure_false: forall
    g g' s ea eb csa csb cfa cfb ma mb
    (CNF: configure w g s false ea eb csa csb cfa cfb ma mb),
  configure w g' s false ea eb csa csb cfa cfb ma mb.
Proof.
  i; inv CNF.
Qed.

Lemma configure_mon: forall g s sync ea eb csa csb cfa cfb ma mb
    (CNF: configure w g s sync ea eb csa csb cfa cfb ma mb)
    g' (LE: g <g= g'),
  configure w g' s sync ea eb csa csb cfa cfb ma mb.
Proof.
  i; inv CNF; econs; destruct sync; eauto using confrel_mon.
Qed.

Lemma vclos__mon sts (mg: method_query La Lb sts): forall s g vr
     va vb (REL: vclos_ mg s g vr  va vb)
    vr' (LE: vr <r= vr'),
  vclos_ mg s g vr'  va vb.
Proof.
  i;  inv REL; eauto.
Qed.

Lemma vclos_induction sts (mg: method_query La Lb sts): forall s g
     va vb (EL: vclos mg s g  va vb)
    P (PF: forall  va vb 
             (IH: vclos_ mg s g P  va vb), 
           P  va vb),
  P  va vb.
Proof.
   fix 5. 
  intros; apply PF; destruct EL; destruct REL
  ; try econstructor (by eauto; eapply vclos_induction; eauto).
Qed.

Lemma vclos_induction_strong sts (mg: method_query La Lb sts): forall s g
     va vb (EL: vclos mg s g  va vb)
    P (PF: forall  va vb
             (IH: vclos_ mg s g (fun  va vb => 
                    P  va vb /\ 
                    vclos mg s g  va vb)  va vb), 
           P  va vb),
  P  va vb.
Proof.
  intros.
  cut (P  va vb /\ vclos mg s g  va vb).
    by intuition.
  revert P PF; eapply (vclos_induction EL); clear  va vb EL.
  i; split; eauto using vclos__mon.
  econstructor; eapply vclos__mon; eauto.
  by simpl; intros; hdes; eauto.
Qed.

Lemma vclos_thunk_mon sts (mg: method_query La Lb sts): forall s g 
     va vb (VC: vclos_thunk mg s g  va vb)
    s' (EXT: sts.(ext) s' s)
    g' (LT: g <r= g'),
  vclos_thunk mg s' g'  va vb.
Proof. i; destruct VC; eauto using vquerya_mon, vqueryb_mon. Qed.

Lemma vclos_mon sts (mg: method_query La Lb sts): forall s g 
     va vb (VC: vclos mg s g  va vb)
    s' (EXT: sts.(ext) s' s)
    g' (LT: g <r= g'),
  vclos mg s' g'  va vb.
Proof.
  intros until vb; intros REL; eapply (vclos_induction REL); intros.
  econs; destruct IH; simpls
  ; econstructor (
      by eauto using vquerya_mon, vqueryb_mon, rquery_mon, vclos_thunk_mon).
Qed.

Lemma lsim_mon: forall (p: esimtyp w) s g  va vb
    (R: lsim_ p g s  va vb)
    p' (LEX: p <e= p')
    g' (LEG: g <g= g')
    s' (EXT: w.(ext) s' s),
  lsim_ p' g' s'  va vb.
Proof. 
  i; destruct R; econs. 
  - destruct WF; eauto 7 using vquerya_mon, vqueryb_mon.
  - i; eapply LEX, CS; eauto using ext_trans.
Qed.

Lemma ksim_mon: forall (p: esimtyp w)
  i kia kib g si sp s  ka kb
    (R: ksim_ p i kia kib g si sp s  ka kb)
    p' (LEX: p <e= p')
    g' (LEG: g <g= g')
    sp' (EXTP: w.(extpub) sp' sp)
    s' (EXT: w.(ext) s' s),
  ksim_ p' i kia kib g' si sp' s'  ka kb.
Proof. 
  econs; ii; destruct R
  ; eauto 10 using extpub_trans, ext_trans.
Qed.

Lemma call_mon: forall s rf rv rk  ea eb
    (STK: call w s rf rv rk  ea eb)
    rf' (LEF: forall  va vb 
              (PR: rf  va vb), rf'  va vb : Prop)
    rv' (LEV: rv <r= rv')
    rk' (LEK: rk <ks= rk'),
  call w s rf' rv' rk'  ea eb.
Proof. 
  i; depdes STK 
  ; econstructor (by eauto using vclos_mon; depdes RF
                  ; econstructor (by eauto; apply LEF; s; eauto)).
Qed.

Lemma gknow_ant: forall es g
    (GK: gknow_ w es g)
    es' (LEE: es' <e= es),
  gknow_ w es' g.
Proof.
  i; destruct GK; econs; eauto using lsim_mon, vclos_thunk_mon.
Qed.

Lemma esim_progress_mon: forall es i kia kib 
  g si s sync evt mb mb'  ea eb
    (R: @esim_progress La Lb w es i kia kib 
          g si s sync evt mb mb'  ea eb)
    es' (LEX: es <e= es')
    i' (GEI: wfo_le i i'),
  esim_progress es' i' kia kib g si s sync evt mb mb'  ea eb.
Proof. 
  i; destruct R; econs
  ; eauto 7 using call_mon, ksim_mon
  ; des; eauto
  ; rdes GEI; subst; esplits; eauto using tc_rtc_tc.
Qed.

Lemma esim_call_mon: forall es kia kib 
  g si s sync evt mb mb'  ea eb
    (R: @esim_call La Lb w es kia kib 
          g si s sync evt mb mb'  ea eb)
    es' (LEX: es <e= es'),
  esim_call es' kia kib g si s sync evt mb mb'  ea eb.
Proof. 
  i; destruct R; econs
  ; eauto 7 using call_mon, ksim_mon
  ; des; eauto
  ; rdes GEI; subst; esplits; eauto using t_step, t_trans.
Qed.

Lemma esim_main_mon: forall esimstep
  kia kib g si s  cfa ma cfb mb
    (R: @esim_main La Lb w esimstep kia kib 
          g si s  cfa ma cfb mb)
    esimstep' (LES: esimstep  <7=  esimstep')
    g' (LE: g <g= g')
    s' (EXT: w.(ext) s s'),
  esim_main esimstep' kia kib g' si s'  cfa ma cfb mb.
Proof.
  i; destruct R.
  - eauto.
  - eapply esim_step; eauto.
    i; specialize (EST _ _ ST); des; subst; des; esplits
    ; eauto using vclos_mon, confrel_mon, ext_trans, configure_mon.
  - eapply esim_ret
    ; eauto using vclos_mon, confrel_mon, ext_trans, configure_mon.
Qed.

Lemma esim_mon: 
   monotone9  (@esim_ La Lb w).
Proof.
  ii; destruct IN; econs; eauto.
  ii; exploit ESIM; eauto using gknow_ant. 
  intro MAIN; des; esplits; eauto.
  eapply esim_main_mon; eauto.
  s; i; des; eauto using esim_progress_mon, esim_call_mon.
Qed.

Lemma gres_mon: 
   monotone9  (@gres La Lb w).
Proof. ii; destructs IN RES; exists clo; eauto. Qed.

Lemma esim_gres_mon: 
   monotone9  (esim_ <*> @gres La Lb w).
Proof.
  ii; eapply esim_mon; eauto.
  i; eapply gres_mon; eauto.
Qed.

Lemma pesim_mon: 
   monotone9  (@pesim La Lb w).
Proof.
  apply paco9_mon.
Qed.

End Simple_Properties_of_Model.

Hint Resolve esim_mon: paco.
Hint Resolve esim_gres_mon: paco.

































































