(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program Relations.
Require Import countable vgtac basis lang_common lang_spec lang_mid.
Require Import model_common model_untyped model_properties_untyped gw_untyped mm.
Require Import ssr util lang_mid_renaming.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".


(** * Expressions annotated with data of type [A]. *)

Inductive aexp (A: Type) :=
| IEunit
| IEnat  (n:nat)
| IEpair (x:itermvar) (y:itermvar)
| IEinl  (x:itermvar)
| IEinr  (x:itermvar)
| IEbop  (op: binop) (x:itermvar + nat) (y:itermvar + nat)
| IEfst  (x:itermvar)
| IEsnd  (x:itermvar)
| IEfix  (bf:termvar) (bx:termvar) (bk:contvar) (e:aexp A)
| IEgen  (bk:contvar) (e:aexp A)
| IEcont (bx:termvar) (e:aexp A)
| IEnew    (x:itermvar) (bx:termvar) (e:aexp A)
| IEread   (x:itermvar) (bx:termvar) (e:aexp A)
| IEwrite  (x:itermvar) (y:itermvar) (e:aexp A)
| IEinput  (bx:termvar) (e:aexp A)
| IEoutput (x:itermvar) (e:aexp A)
| IEifnz   (x:itermvar) (ea:aexp A) (eb:aexp A)
| IEcase   (x:itermvar) (bx:termvar) (ea:aexp A) (eb:aexp A)
| IEret    (xk:contvar) (x:itermvar)
| IEapp (a:A) (xf:itermvar) (x:itermvar) (xk:contvar)
| IEinst (a:A) (xf:itermvar) (xk:contvar)
| IElet (a:A) (ae:aexp A) (bx:termvar + contvar) (e:aexp A)
| IErefeq (x:itermvar) (y:itermvar) 
.


(** Erase annotations. *)

Fixpoint aerase A (e: aexp A) : exp :=
  match e with
    | IEunit
      => lang_mid.IEunit
    | IEnat n
      => lang_mid.IEnat n
    | IEpair x y
      => lang_mid.IEpair x y
    | IEret k x
      => lang_mid.IEret k x
    | IEinst _ x k
      => lang_mid.IEinst x k
    | IEinl x
      => lang_mid.IEinl x
    | IEinr x
      => lang_mid.IEinr x
    | IEfst x
      => lang_mid.IEfst x
    | IEsnd x
      => lang_mid.IEsnd x
    | IEbop b xn yn
      => lang_mid.IEbop b xn yn
    | IEapp _ x y k
      => lang_mid.IEapp x y k
    | IErefeq x y
      => lang_mid.IErefeq x y

    | IEfix f x k ea
      => lang_mid.IEfix f x k (aerase ea)
    | IEgen x ea
      => lang_mid.IEgen x (aerase ea)
    | IEcont x ea
      => lang_mid.IEcont x (aerase ea)
    | IEnew x y ea
      => lang_mid.IEnew x y (aerase ea)
    | IEread x y ea
      => lang_mid.IEread x y (aerase ea)
    | IEoutput x ea
      => lang_mid.IEoutput x (aerase ea)
    | IEwrite x y ea
      => lang_mid.IEwrite x y (aerase ea)
    | IEinput x ea
      => lang_mid.IEinput x (aerase ea)

    | IEifnz x ea eb
      => lang_mid.IEifnz x (aerase ea) (aerase eb)
    | IEcase x y ea eb
      => lang_mid.IEcase x y (aerase ea) (aerase eb)
    | IElet t ea x eb 
      => lang_mid.IElet (aerase ea) x (aerase eb)
  end.


Definition annotate_with A a : exp -> aexp A :=
  fix aw (e: exp) :=
  match e with
    | lang_mid.IElet e1 x e2
      => IElet a (aw e1) x (aw e2)

    | lang_mid.IEapp x y k
      => IEapp a x y k

    | lang_mid.IEinst x k
      => IEinst a x k

    | lang_mid.IEunit
      => IEunit _
    | lang_mid.IEnat n
      => IEnat _ n
    | lang_mid.IEpair x y
      => IEpair _ x y
    | lang_mid.IEret k x
      => IEret _ k x
    | lang_mid.IEinl x
      => IEinl _ x
    | lang_mid.IEinr x
      => IEinr _ x
    | lang_mid.IEfst x
      => IEfst _ x
    | lang_mid.IEsnd x
      => IEsnd _ x
    | lang_mid.IEbop b xn yn
      => IEbop _ b xn yn
    | lang_mid.IErefeq x y
      => IErefeq _ x y

    | lang_mid.IEfix f x k ea
      => IEfix f x k (aw ea)
    | lang_mid.IEgen x ea
      => IEgen x (aw ea)
    | lang_mid.IEcont x ea
      => IEcont x (aw ea)
    | lang_mid.IEnew x y ea
      => IEnew x y (aw ea)
    | lang_mid.IEread x y ea
      => IEread x y (aw ea)
    | lang_mid.IEoutput x ea
      => IEoutput x (aw ea)
    | lang_mid.IEwrite x y ea
      => IEwrite x y (aw ea)
    | lang_mid.IEinput x ea
      => IEinput x (aw ea)

    | lang_mid.IEifnz x ea eb
      => IEifnz x (aw ea) (aw eb)
    | lang_mid.IEcase x y ea eb
      => IEcase x y (aw ea) (aw eb)
  end.

Lemma annotate_with_erase: forall A (a: A) e,
  aerase (annotate_with a e) = e.
Proof.
  i; induction e; s; rewrite ?IHe ?IHe1 ?IHe2; eauto.
Qed.



(** * Expression transformations as annotations. *)

Definition exptf :=
  exp -> option exp.

Definition exptf_wf f :=
  forall e e' G
      (E: f e = Some e')
      (WF: wfexp G e)
      (UN: uniq (bvars e ++ purge_lbl G)),
    <<WF: wfexp G e'>> /\
    <<BV: subuniq (bvars e') (bvars e)>> /\
    <<GT: is_good_thunk e -> is_good_thunk e'>> /\
    <<GE: good_exp e -> good_exp e'>>.

Definition exptf_sim f :=
  forall e e' G lw
      (E: f e = Some e')
      (WF: wfexp G e)
      (UN: uniq (bvars e ++ purge_lbl G))
      (ST: stable lw),
    rtc (mm_psim lw G ren_id) e' e.

Record annT :=
  { _annT :> Type
  ; perform: _annT -> exptf
  ; perform_wf: forall a, exptf_wf (perform a)
  ; perform_sim: forall a, exptf_sim (perform a)
  }.
Arguments perform [a] _ _.


(** ** Sum lifting for annotation types. *)

Definition annT_sum (A B: annT) : annT.
  refine {| _annT := A + B
          ; perform ann := match ann with
                             | inl a => perform a
                             | inr b => perform b
                           end
         |}.
  - i; destruct a; eauto using perform_wf.
  - i; destruct a; eauto using perform_sim.
Defined.


(** ** List lifting for annotation types.

    Transformations will be [perform]ed from right to left until one
    fails, in which case its input is returned.

    (We may want to also have list liftings with a different
    semantics.)
*)


(** *** Raw definition. *)

Fixpoint perform_list' (A: annT) (revanns: list A) : exptf :=
  fun e => match revanns with
             | nil => Some e
             | a::revanns' => match perform a e with
                                | None => Some e
                                | Some e' => perform_list' A revanns' e'
                              end
  end.

Definition perform_list (A: annT) (anns: list A) : exptf :=
  perform_list' A (rev anns).


(** *** Syntactic correctness. *)

Lemma perform_list_wf: forall A anns,
  exptf_wf (perform_list A anns).
Proof.
  i; unfold perform_list; generalize (rev anns); clear anns.
  induction l; ii; s in E; clarify; [by splits; auto|].
  destruct (perform a e) eqn:E'.
  - exploit (perform_wf A); eauto; []; i; des.
    exploit IHl; eauto using subuniq_uniq_app; []; i; des.
    esplits; eauto using subuniq_trans.
  - clarify; splits; auto.
Qed.


(** *** Semantic correctness. *)

Lemma perform_list_sim: forall A anns,
  exptf_sim (perform_list A anns).
Proof.
  i; unfold perform_list; generalize (rev anns); clear anns.
  induction l; ii; s in E; clarify; [by splits; auto|].
  destruct (perform a e) eqn:E'.
  - exploit (perform_wf A); eauto; []; i; des.
    exploit (perform_sim A); eauto; []; i; des.
    exploit IHl; eauto using subuniq_uniq_app; []; i; des.
    esplits; eauto using subuniq_trans.
    econs 3; eauto.
  - clarify; splits; auto.
Qed.


(** *** Final definition. *)

Definition annT_list (A: annT) : annT :=
  {| _annT := list A
   ; perform := perform_list _
   ; perform_wf := perform_list_wf _
   ; perform_sim := perform_list_sim _
  |}.



(** ** Process all annotations of an expression. *)


(** *** Definition. *)

Definition transform {A: annT} : aexp A -> exp :=
  fix transform e :=
    match e with
      | IElet a ea x eb 
        => let e' := lang_mid.IElet (transform ea) x (transform eb) in
           withd e' (perform a e')

      | IEapp a x y k
        => let e' := lang_mid.IEapp x y k in
           withd e' (perform a e')

      | IEinst a x k
        => let e' := lang_mid.IEinst x k in
           withd e' (perform a e')

      | IEfix f x k ea
        => lang_mid.IEfix f x k (transform ea)
      | IEgen k ea
        => lang_mid.IEgen k (transform ea)
      | IEcont x ea
        => lang_mid.IEcont x (transform ea)
      | IEnew x y ea
        => lang_mid.IEnew x y (transform ea)
      | IEread x y ea
        => lang_mid.IEread x y (transform ea)
      | IEoutput x ea
        => lang_mid.IEoutput x (transform ea)
      | IEwrite x y ea
        => lang_mid.IEwrite x y (transform ea)
      | IEinput x ea
        => lang_mid.IEinput x (transform ea)
      | IEifnz x ea eb
        => lang_mid.IEifnz x (transform ea) (transform eb)
      | IEcase x y ea eb
        => lang_mid.IEcase x y (transform ea) (transform eb)

      | IEunit
        => lang_mid.IEunit
      | IEnat n
        => lang_mid.IEnat n
      | IEpair x y
        => lang_mid.IEpair x y
      | IEret k x
        => lang_mid.IEret k x
      | IEinl x
        => lang_mid.IEinl x
      | IEinr x
        => lang_mid.IEinr x
      | IEfst x
        => lang_mid.IEfst x
      | IEsnd x
        => lang_mid.IEsnd x
      | IEbop b xn yn
        => lang_mid.IEbop b xn yn
      | IErefeq x y
        => lang_mid.IErefeq x y
    end.

Definition transform_mod {A: annT} (analyze: exp -> aexp A) : module -> module :=
  modmap (@transform A <*> analyze).


(** *** Syntactic correctness. *)

Lemma transform_wfatom_: forall (A: annT) e (ae: aexp A) G
    (WF: wfatom G e)
    (UN: uniq (bvars e ++ purge_lbl G))
    (E: aerase ae = e)
    (IH: forall e', size e' < size e -> forall ae' G, wfexp G e' ->
           uniq (bvars e' ++ purge_lbl G) -> @aerase A ae' = e' ->
           <<WF: wfexp G (transform ae')>> /\
           <<BV: subuniq (bvars (transform ae')) (bvars e')>> /\
           <<GT: good_exp e' -> good_exp (transform ae')>>),
  <<WF: wfatom G (transform ae)>> /\
  <<BV: subuniq (bvars (transform ae)) (bvars e)>> /\
  <<GT: is_good_thunk e -> is_good_thunk (transform ae)>>.
Proof.
  i; destruct ae; try by splits; vauto.
  - s in E; hexploit (IH (aerase ae)); vauto.
    { 
      simpl bvars in UN; eapply uniq_sublist; [eapply uniq_perm; first by eauto|].
      - rewrite cons_app3; eapply Permutation_app; [|by eauto].
        eapply perm_trans; [by eapply Permutation_app_comm|].
        eapply Permutation_app; [by eauto|]; eapply Permutation_rev.
      - rewrite -app_assoc; apply sublist_app; auto.
        s; eauto using purge_lbl_sublist.
    }
    i; des; splits; vauto; repeat econs 2; eauto.
    + simpl bvars; rewrite cons_app3; rhs rewrite cons_app3.
      eauto using subuniq_app.
    + s; intros; apply H; invc FH; rewrite -H4 in GT.
      exfalso; eapply GT; destruct ae; vauto.
  - s in E; hexploit (IH (aerase ae)); vauto.
    { simpl bvars in UN; eapply uniq_sublist; [eapply uniq_perm; first by eauto|].
      - rewrite cons_app; eapply Permutation_app; [|by eauto].
        eapply perm_trans; [by eapply Permutation_app_comm|].
        eapply Permutation_app; [by eauto|]; eapply Permutation_rev.
      - rewrite -app_assoc; apply sublist_app; auto.
        s; eauto using purge_lbl_sublist.
    }
    i; des; splits; vauto; repeat econs 2; eauto.
    + simpl bvars; rewrite cons_app; rhs rewrite cons_app.
      eauto using subuniq_app.
    + s; intros; apply H; invc FH; rewrite -H2 in GT.
      exfalso; eapply GT; destruct ae; vauto.
  - s in E; hexploit (IH (aerase ae)); vauto.
    { simpl bvars in UN; eapply uniq_sublist; [eapply uniq_perm; first by eauto|].
      - rewrite cons_app; eapply Permutation_app; [|by eauto].
        eapply perm_trans; [by eapply Permutation_app_comm|].
        eapply Permutation_app; [by eauto|]; eapply Permutation_rev.
      - rewrite -app_assoc; apply sublist_app; auto.
    }
    i; des; splits; vauto; repeat econs 2; eauto.
    simpl bvars; rewrite cons_app; rhs rewrite cons_app.
    eauto using subuniq_app.
Qed.


Lemma transform_wfexp: forall (A: annT) e (ae: aexp A) G
    (E: aerase ae = e)
    (WF: wfexp G e)
    (UN: uniq (bvars e ++ purge_lbl G)),
  <<WF: wfexp G (transform ae)>> /\
  <<BV: subuniq (bvars (transform ae)) (bvars e)>> /\
  <<GE: good_exp e -> good_exp (transform ae)>>.
Proof.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  i; destruct ae; vauto; try (by s in WF; des; s; splits; eauto).
  - s in WF; des.
    hexploit (IH (aerase ae)); vauto.
    { simpl bvars in UN; rewrite -app_comm_cons in UN.
      eapply uniq_perm; eauto using Permutation_cons_app, purge_lbl.
    }
    i; des; splits; vauto.
    simpl bvars; rewrite cons_app; rhs rewrite cons_app.
    eauto using subuniq_app.
  - s in WF; des.
    hexploit (IH (aerase ae)); vauto.
    { simpl bvars in UN; rewrite -app_comm_cons in UN.
      eapply uniq_perm; eauto using Permutation_cons_app.
    }
    i; des; splits; vauto.
    simpl bvars; rewrite cons_app; rhs rewrite cons_app.
    eauto using subuniq_app.
  - s in WF; des.
    hexploit (IH (aerase ae)); vauto.
    i; des; splits; vauto.
  - s in WF; des.
    hexploit (IH (aerase ae)); vauto.
    { simpl bvars in UN; rewrite -app_comm_cons in UN.
      eapply uniq_perm; eauto using Permutation_cons_app.
    }
    i; des; splits; vauto.
    simpl bvars; rewrite cons_app; rhs rewrite cons_app.
    eauto using subuniq_app.
  - s in WF; des.
    hexploit (IH (aerase ae)); vauto.
    i; des; splits; vauto.
  - s in WF; des.
    hexploit (IH (aerase ae1)); first (by rr; s; omega); vauto.
    { simpl bvars in UN; rewrite -app_assoc in UN.
      eapply uniq_sublist; eauto using sublist_app.
    }
    hexploit (IH (aerase ae2)); first (by rr; s; omega); vauto.
    { simpl bvars in UN; rewrite -app_assoc in UN.
      eapply uniq_sublist; eauto using sublist_app.
    }
    i; des; splits; vauto.
    simpl bvars; eauto using subuniq_app.
  - s in WF; des.
    hexploit (IH (aerase ae1)); first (by rr; s; omega); vauto.
    { simpl bvars in UN; rewrite -app_comm_cons in UN.
      eapply uniq_perm; last by eapply Permutation_cons_app; eauto.
      eapply uniq_sublist; [by eauto|].
      econs 3; rewrite -app_assoc; eauto using sublist_app.
    }
    hexploit (IH (aerase ae2)); first (by rr; s; omega); vauto.
    { simpl bvars in UN; rewrite -app_comm_cons in UN.
      eapply uniq_perm; last by eapply Permutation_cons_app; eauto.
      eapply uniq_sublist; [by eauto|].
      econs 3; rewrite -app_assoc; eauto using sublist_app.
    }
    i; des; splits; vauto.
    simpl bvars; rewrite cons_app; rhs rewrite cons_app.
    eauto using subuniq_app.
  - s in WF; des; s.
    destruct (perform _ _) eqn:E; simpl withd.
    + exploit perform_wf; eauto; [by econs; eauto|]; i; des.
      splits; eauto.
    + splits; eauto; econs; eauto.
  - s in WF; des; s.
    destruct (perform _ _) eqn:E; simpl withd.
    + exploit perform_wf; eauto; [by econs; eauto|]; i; des.
      splits; eauto.
    + splits; eauto; econs; eauto.
  - s in WF; destruct bx; des.
    + simpl transform.
      hexploit transform_wfatom_; eauto.
      { simpl bvars in UN; rewrite -app_comm_cons in UN.
        eapply uniq_sublist; eauto using sublist_app.
      }
      { i; eapply IH; auto; rr in H; rr; s; omega. }
      hexploit (IH (aerase ae2)); first (by rr; s; omega); vauto.
      { simpl bvars in UN; rewrite -app_comm_cons in UN.
        eapply uniq_perm; last by eapply Permutation_cons_app; eauto.
        eapply uniq_sublist; [by eauto|].
        econs 3; rewrite -app_assoc; eauto using sublist_app.
      }
      i; des; s; destruct (perform _ _) eqn:E; simpl withd; last first.
      { splits; s; eauto using Permutation_app; vauto.
        rewrite cons_app; rhs rewrite cons_app; eauto using subuniq_app.
      }
      exploit perform_wf; eauto.
      * s; eauto.
      * simpl bvars; simpl bvars in UN.
        eapply subuniq_uniq_app; eauto.
        rewrite cons_app; rhs rewrite cons_app; eauto using subuniq_app.
      * i; des; splits; eauto; [|by ii; rewrite H0 in GE0; eapply GE0; vauto].
        s; s in BV1; eapply subuniq_trans; eauto.
        rewrite cons_app; rhs rewrite cons_app; eauto using subuniq_app.
    + simpl transform.
      hexploit transform_wfatom_; eauto.
      { simpl bvars in UN; rewrite -app_comm_cons in UN.
        eapply uniq_sublist; eauto using sublist_app.
      }
      { i; eapply IH; auto; rr in H; rr; s; omega. }
      hexploit (IH (aerase ae2)); first (by rr; s; omega); vauto.
      { simpl bvars in UN; rewrite -app_comm_cons in UN.
        eapply uniq_perm; last by eapply Permutation_cons_app; eauto.
        eapply uniq_sublist; [by eauto|].
        econs 3; rewrite -app_assoc; eauto using sublist_app.
      }
      i; des; s; destruct (perform _ _) eqn:E; simpl withd; last first.
      { splits; s; eauto using Permutation_app; vauto.
        rewrite cons_app; rhs rewrite cons_app; eauto using subuniq_app.
      }
      exploit perform_wf; eauto.
      * s; eauto.
      * simpl bvars; simpl bvars in UN.
        eapply subuniq_uniq_app; eauto.
        rewrite cons_app; rhs rewrite cons_app; eauto using subuniq_app.
      * i; des; splits; eauto; [|by ii; rewrite H0 in GE0; eapply GE0; vauto].
        s; s in BV1; eapply subuniq_trans; eauto.
        rewrite cons_app; rhs rewrite cons_app; eauto using subuniq_app.
Qed.

Lemma transform_wfatom: forall (A: annT) e (ae: aexp A) G
    (E: aerase ae = e)
    (WF: wfatom G e)
    (UN: uniq (bvars e ++ purge_lbl G)),
  <<WF: wfatom G (transform ae)>> /\
  <<BV: subuniq (bvars (transform ae)) (bvars e)>> /\
  <<GT: is_good_thunk e -> is_good_thunk (transform ae)>>.
Proof.
  i; apply transform_wfatom_; auto.
  i; apply transform_wfexp; auto.
Qed.

Theorem transform_wfmod: forall (A: annT) (analyze: exp -> aexp A) FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m)
    (AN: forall e, aerase (analyze e) = e),
  mid_wfmod FI (transform_mod analyze m) FO /\
  uniqvar (transform_mod analyze m).
Proof.
  i; apply modmap_wfmod; eauto; []; i.
  dup UN; apply uniq_appD in UN; des.
  hexploit transform_wfatom; eauto.
  - rewrite purge_lbl_map_inl_inl; rewrite app_nil_r; eauto.
  - i; des; splits; eauto using subuniq_uniq_app.
Qed.


(** *** Semantic correctness. *)

Ltac uniq_auto H :=
  try (by simpl bvars in H;
       eapply uniq_sublist; [eapply uniq_perm; first by eauto|]; inst;
       [rewrite ->cons_app, <-app_assoc; eapply Permutation_cons_app; eauto
       |rewrite -app_assoc; eauto using sublist_app]).


Lemma transform_asim: forall (A: annT) (ae: aexp A) G e
    (E: aerase ae = e)
    (WF: wfatom G e)
    (UN: uniq (bvars e ++ purge_lbl G))
    (IH: forall e', size e' < size e -> forall G ae',
           wfexp G e' -> uniq (bvars e' ++ purge_lbl G) -> @aerase A ae' = e' ->
           rtc (mm_psim (lwemp mm_gw) G ren_id) (transform ae') e'),
  rtc (mm_asim (lwemp mm_gw) G ren_id) (transform ae) e.
Proof.
  i; destruct ae; vauto.
  - simpl transform; simpl aerase.
    eapply mmrtc_compat_fix; eauto.
    eapply IH; eauto.
    simpl bvars in UN; eapply uniq_sublist; [eapply uniq_perm; first by eauto|].
    + rewrite cons_app3; eapply Permutation_app; [|by eauto].
      eapply perm_trans; [by eapply Permutation_app_comm|].
      eapply Permutation_app; [by eauto|]; eapply Permutation_rev.
    + simpl; rewrite cons_app3 -app_assoc.
      repeat eapply sublist_app; eauto.
      s; eauto using purge_lbl_sublist.
  - simpl transform; simpl aerase.
    eapply mmrtc_compat_gen; eauto.
    eapply IH; eauto.
    simpl bvars in UN; eapply uniq_sublist; [eapply uniq_perm; first by eauto|].
    + rewrite cons_app; eapply Permutation_app; [|by eauto].
      eapply perm_trans; [by eapply Permutation_app_comm|].
      eapply Permutation_app; [by eauto|]; eapply Permutation_rev.
    + simpl; rewrite cons_app -app_assoc.
      repeat eapply sublist_app; eauto.
      s; eauto using purge_lbl_sublist.
Qed.

Lemma transform_ksim: forall (A: annT) (ae: aexp A) G e
    (E: aerase ae = e)
    (WF: wfatom G e)
    (UN: uniq (bvars e ++ purge_lbl G))
    (IH: forall e', size e' < size e -> forall G ae',
           wfexp G e' -> uniq (bvars e' ++ purge_lbl G) -> @aerase A ae' = e' ->
           rtc (mm_psim (lwemp mm_gw) G ren_id) (transform ae') e'),
  rtc (mm_ksim (lwemp mm_gw) G ren_id) (transform ae) e.
Proof.
  i; destruct ae; vauto; s in WF.
  hexploit (IH (aerase ae)); eauto.
  { simpl bvars in UN; eapply uniq_perm; eauto using Permutation_cons_app. }
  i; des; simpl transform; simpl aerase.
  eapply mmrtc_compat_cont; eauto.
Qed.

Lemma transform_psim: forall (A: annT) e (ae: aexp A) G
    (E: aerase ae = e)
    (WF: wfexp G e)
    (UN: uniq (bvars e ++ purge_lbl G)),
  rtc (mm_psim (lwemp mm_gw) G ren_id) (transform ae) e.
Proof.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  i; destruct ae; i; vauto.
  - s; s in WF; des; eapply mmrtc_compat_new; eauto using lwemp_stable.
    eapply (IH (aerase ae)); vauto.
    eapply uniq_perm; eauto; inst; s; eauto using Permutation_cons_app.
  - s; s in WF; des; eapply mmrtc_compat_read; eauto using lwemp_stable.
    eapply (IH (aerase ae)); vauto.
    eapply uniq_perm; eauto; inst; s; eauto using Permutation_cons_app.
  - s; s in WF; des; eapply mmrtc_compat_write; eauto using lwemp_stable.
  - s; s in WF; des; eapply mmrtc_compat_input; eauto using lwemp_stable.
    eapply (IH (aerase ae)); vauto.
    eapply uniq_perm; eauto; inst; s; eauto using Permutation_cons_app.
  - s; s in WF; des; eapply mmrtc_compat_output; eauto using lwemp_stable.
  - s; s in WF; des; eapply mmrtc_compat_ifnz; eauto using lwemp_stable.
    + eapply (IH (aerase ae1)); auto; [rr; s; omega|].
      s in UN; eapply uniq_sublist; eauto.
      rewrite -app_assoc; eauto using sublist_app.
    + eapply (IH (aerase ae2)); auto; [rr; s; omega|].
      simpl bvars in UN; eapply uniq_sublist; eauto.
      rewrite -app_assoc; eauto using sublist_app.
  - s; s in WF; des; eapply mmrtc_compat_case; eauto using lwemp_stable.
    + eapply (IH (aerase ae1)); auto; [rr; s; omega|].
      simpl bvars in UN.
      eapply uniq_sublist; [eapply uniq_perm; first by eauto|].
      * rewrite ->cons_app, <-app_assoc; eapply Permutation_cons_app; eauto.
      * rewrite -app_assoc; eauto using sublist_app.
    + eapply (IH (aerase ae2)); auto; [rr; s; omega|].
      simpl bvars in UN.
      eapply uniq_sublist; [eapply uniq_perm; first by eauto|].
      * rewrite ->cons_app, <-app_assoc; eapply Permutation_cons_app; eauto.
      * rewrite -app_assoc; eauto using sublist_app.
  - s; destruct (perform _ _) eqn:E; simpl withd; eauto.
    
    eapply perform_sim; eauto using lwemp_stable
    ; (destruct bx; simpl bvars in UN; simpl bvars; s in WF; des
       ; hexploit transform_wfexp; eauto; uniq_auto UN
       ; hexploit transform_wfatom; eauto; uniq_auto UN
       ; inst; i; des; try (by eauto)
       ; eapply subuniq_uniq_app; eauto
       ; rewrite cons_app; rhs rewrite cons_app; eauto using subuniq_app).
  - s; destruct (perform _ _) eqn:E; simpl withd; eauto.
    
    eapply perform_sim; eauto using lwemp_stable
    ; (destruct bx; simpl bvars in UN; simpl bvars; s in WF; des
       ; hexploit transform_wfexp; eauto; uniq_auto UN
       ; hexploit transform_wfatom; eauto; uniq_auto UN
       ; inst; i; des; try (by eauto)
       ; eapply subuniq_uniq_app; eauto
       ; rewrite cons_app; rhs rewrite cons_app; eauto using subuniq_app).
  - s; destruct (perform _ _) eqn:E; simpl withd.
    + 
      econs 3.
      * s; eapply perform_sim; eauto using lwemp_stable
        ; (destruct bx; simpl bvars in UN; simpl bvars; s in WF; des
           ; hexploit transform_wfexp; eauto; uniq_auto UN
           ; hexploit transform_wfatom; eauto; uniq_auto UN
           ; inst; i; des; try (by eauto)
           ; eapply subuniq_uniq_app; eauto
           ; rewrite cons_app; rhs rewrite cons_app; eauto using subuniq_app).
      * destruct bx; simpl bvars in UN; simpl bvars; s in WF; des
        ; hexploit transform_wfexp; eauto; uniq_auto UN
        ; hexploit transform_wfatom; eauto; uniq_auto UN
        ; inst; i; des; simpl aerase.
        { eapply mmrtc_compat_let; eauto using lwemp_stable.
          - eapply transform_asim; eauto; uniq_auto UN.
            i; eapply IH; eauto; rr; s; omega.
          - eapply (IH (aerase ae2)); eauto; uniq_auto UN; rr; s; omega.
        }
        { eapply mmrtc_compat_letcont; eauto using lwemp_stable.
          - eapply transform_ksim; eauto; uniq_auto UN.
            i; eapply IH; eauto; rr; s; omega.
          - eapply IH; eauto; uniq_auto UN; rr; s; omega.
        }
    + destruct bx; simpl bvars in UN; simpl bvars; s in WF; des
        ; hexploit transform_wfexp; eauto; uniq_auto UN
        ; hexploit transform_wfatom; eauto; uniq_auto UN
        ; inst; i; des; simpl aerase.
      { eapply mmrtc_compat_let; eauto using lwemp_stable.
        - eapply transform_asim; eauto; uniq_auto UN.
          i; eapply IH; eauto; rr; s; omega.
        - eapply (IH (aerase ae2)); eauto; uniq_auto UN; rr; s; omega.
      }
      { eapply mmrtc_compat_letcont; eauto using lwemp_stable.
        - eapply transform_ksim; eauto; uniq_auto UN.
          i; eapply IH; eauto; rr; s; omega.
        - eapply IH; eauto; uniq_auto UN; rr; s; omega.
      }
Qed.

Theorem transform_msim: forall (A: annT) (analyze: exp -> aexp A) FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m)
    (AN: forall e, aerase (analyze e) = e),
  rtc (msim mm_gw FI FO) (transform_mod analyze m) m.
Proof.
  i; ginduction WF; eauto.
  i; apply uniq_appD in UV; fold bvars_mod in UV; des.
  hexploit IHWF; eauto; []; i.
  eapply mm_rtcmsim_cons; eauto using lwemp_stable.
  - hexploit (AN e); [].
    destruct e; vauto.
    + destruct (analyze _) eqn:E; vauto; simpl aerase; i; clarify; [].
      s in WFE; rewrite map_comp purge_map_inl -map_comp in WFE.
      s; rewrite E; s; apply mmrtc_tlsim_fix, transform_psim; eauto.
      eapply uniq_perm; eauto.
      s; rewrite purge_lbl_map_inl_inl cons_app3.
      eapply Permutation_trans; [by eapply Permutation_app_comm|].
      eapply Permutation_app; [by auto|].
      eapply Permutation_trans; [by eapply Permutation_rev|]; eauto.
    + destruct (analyze _) eqn:E; vauto; simpl aerase; i; clarify; [].
      s in WFE; rewrite map_comp purge_map_inl -map_comp in WFE.
      s; rewrite E; s; apply mmrtc_tlsim_gen, transform_psim; eauto.
      eapply uniq_perm; eauto.
      s; rewrite purge_lbl_map_inl_inl cons_app.
      eapply Permutation_trans; [by eapply Permutation_app_comm|].
      eapply Permutation_app; [by auto|].
      eapply Permutation_trans; [by eapply Permutation_rev|]; eauto.
  - inst_pairs; i; rr; s; unfold mid_cemp; eauto.
Grab Existential Variables.
  exact tt.
Qed.
