(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Omega Program ClassicalChoice.
Require Import vgtac basis myclassical lang_common lang_spec countable.
Require Import lang_tgt lang_mid gw_common model_properties_common ssr.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".



(** * Auxiliaries. *)


Lemma and_with_l: forall (A B: Prop),
  A /\ (A -> B) -> A /\ B.
Proof.
  tauto.
Qed.

Lemma and_with_r: forall (A B: Prop),
  (B -> A) /\ B -> A /\ B.
Proof.
  tauto.
Qed.

Lemma and_l_without: forall (A B C: Prop),
  A /\ (C -> B) -> C -> A /\ B.
Proof.
  tauto.
Qed.

Lemma and_r_without: forall (A B C: Prop),
  (C -> A) /\ B -> C -> A /\ B.
Proof.
  tauto.
Qed.


Ltac pmplus_eq' :=
  match goal with |- ?x ## _ = _ =>
    rhs rewrite <-?pmplus_assoc, <-?(pmplus_comm x), <-?(pmplus_swap x)
  end; eauto.

Ltac pmplus_eq :=
  let X := fresh "newloc" in try set (X := alloc _ _);
  rewrite ->pmplus_comm;
  pmplus_eq'; try unfold X; eauto.

Ltac pmplus_cell_eq :=
  match goal with |- {?x |-> _} ## _ = _ =>
    rhs rewrite ->?pmcells_cons, <-?pmplus_assoc, <-?(pmplus_swap {x |-> _})
  end; eauto.                   


Ltac pmplus_eq'' :=
  try (match goal with |- ?x ## _ = _ =>
    rhs rewrite <-(pmplus_swap x)
  end); inst;
  let y := fresh in 
  try (match goal with |- ?x ## _ = _ =>
    set (y := x)
  end); inst;
  pmplus_eq';
  unfold y in *.


(** * Utilities specific to target language. *)


(** Saves you from also having to show validity and finiteness of the
    smaller memories. *)
Lemma tgt_mcplus: forall m c mc (c': conf tgt_lsi)
    (EC: c' = {| cpc := Just (mpc m); crf := Just (mrf m);
                 cst := mst m; chp := mhp m |})
    (RL: real _ (c' # c) mc),
  mcplus tgt_lsi m c mc.
Proof.
  i; subst; rrdes RL.
  destruct m, c, mc; ss; subst.
  destruct cpc, crf; vauto.
  dup FIN; dup HV; dup SV; split_pmfinite FIN HV; split_pmvalid SV.
  rr; eapply (ex_intro _ (mkconf _ _ _ _)); ss; splits; rr; esplits; s; eauto.
Qed.


(** Splitting the continuation stack. *)
Lemma stack_split: forall i n (cur: list word)
    (IN: i < n)
    (LEN: length cur = 2 * n),
  exists popped used,
    <<SPL: cur = used ++ popped>> /\
    <<POPPED: popped = skipn (length cur- 2 * (1 + i)) cur>> /\
    <<USED: used = firstn (length cur - 2 * (1 + i)) cur>> /\
    <<LU: length used = length cur - 2 * (1 + i)>> /\
    <<LF: length popped = 2 * (1 + i)>>.
Proof.
  i.
  assert (SPL := firstn_skipn (2 * n - 2 * (1 + i)) cur).
  set (popped := skipn (2 * n - 2 * (1 + i)) cur).
  set (used := firstn (2 * n - 2 * (1 + i)) cur).
  assert (length cur = length (used ++ popped)) by (rewrite -SPL; auto).
  assert (length used = 2 * n - 2 * (1 + i)) by (rewrite firstn_length min_l; omega).
  rewrite LEN; esplits; eauto.
  rewrite app_length in H; unfold used, popped in *; rewrite H0 in H; omega.
Qed.


(** Useful when about to apply pesim_release. *)
Lemma tgt_csplit: forall pc rf st hp,
     {| cpc := Just pc; crf := Just rf; cst := st; chp := hp |} =
     @cplus tgt_lsi {| cpc := Just pc; crf := Nothing; cst := pmemp; chp := pmemp |}
                    {| cpc := Nothing; crf := Just rf; cst := st; chp := hp |}.
Proof.
  i; s; noemp; auto.
Qed.


(** List predicate: [x] represents linked list [ys] in heap [h] (intuitionistic).

    Note that this definition requires two cells even for the last
    entry, which is not ideal but simple.
*)
Fixpoint reprlist h x ys : Prop :=
  match ys with
    | nil => True
    | y :: ys' => exists x', pmread h x y /\ pmread h (1 + x) x' /\
                             reprlist h x' ys'
  end.

Lemma reprlist_mon: forall h h' x ys
    (ENV: reprlist h x ys)
    (SH: forall w v (RD: pmread h w v), pmread h' w v),
  reprlist h' x ys.
Proof.
  i; ginduction ys; auto; i; ss; des.
  esplits; eauto.
Qed.


(** Exposing free stack cells. *)
Lemma tgt_stack: forall s c (CP: tgt_confpred s c) n,
  exists vs freest,
    <<NST: length vs = n>> /\
    <<FST: cst c = pmcells (fst s r_sp) vs ## freest>>.
Proof.
  i; rrdes CP; induction n; [by ex @nil; esplits; unfold pmcells; noemp|].
  des; assert (ST' := ST (n + fst s r_sp)); des.
  exploit ST'; [omega|]; i; des; rrdes x0; subst.
  rewrite FST in RD, VLD; eapply pmcells_distinct in RD; [|by auto|omega].
  des; subst; pmcells_foldH FST; rewrite pmcells_sng in FST.
  rewrite ->pmplus_assoc, plus_comm, <-pmcells_app in FST.
  esplits; eauto; inst; clear; induction vs; s; eauto.
Qed.


(** Extending a database. *)
Definition extend X Y (S: X -> Y -> Prop) x y :=
  fun x' y' => x = x' /\ y = y' \/ S x' y'.


Ltac setnewloc :=
  repeat match goal with
    | _ : context[alloc ?a ?b] |- _ =>
      let X := fresh "newloc" in set (X := alloc a b) in *
    | |- context[alloc ?a ?b] =>
      let X := fresh "newloc" in set (X := alloc a b) in *
  end.

Ltac destruct_tgt_real RL :=
  dup RL; destruct RL as (PC & RF & ST & SV & HP & HV & FIN); unnw;
  unfold lang_tgt.cpc, lang_tgt.mpc in PC;
  unfold lang_tgt.crf, lang_tgt.mrf in RF;
  unfold lang_tgt.cst, lang_tgt.mst in SV, ST;
  unfold lang_tgt.chp, lang_tgt.mhp in HV, FIN, HP;
  clean; noemp; subst.


Notation ts_rf := (fun s => fst (fst (fst (fst s)))).
Notation ts_db := (fun s => snd (fst (fst (fst s)))).
Notation ts_lhp := (fun s  => snd (snd s)).
Notation ts_lst := (fun s  => fst (snd s)).



Tactic Notation "setmrf" ident(id) :=
  match goal with |- context[{|mpc := _; mrf := ?x; mst := _; mhp := _|}] =>
    set (id := x)
  end.

Tactic Notation "setcrf" ident(id) :=
  match goal with |- context[{|cpc := _; crf := Just ?x; cst := _; chp := _|}] =>
    set (id := x)
  end.

















Ltac prep_store :=
  match goal with
    | |- store _ ?rf ({?x |-> _} ## _) _ (o_sind ?r ?o) _ _ =>
      replace x with (add_offset (rf r) o)
      ; last (s; try unfold rf; tmext_simpl; s; omega)
    | |- store _ ?rf _ ({?x |-> _} ## _) (o_hind ?r ?o) _ _ =>
      replace x with (add_offset (rf r) o)
      ; last (s; try unfold rf; tmext_simpl; s; omega)
  end.

Ltac read_code Eladdr :=
  match goal with |- pmread _ ?mpc _ =>
    eapply pmread_plus
    ; [by eapply pmread_cell|by rewrite Eladdr; pmplus_cell_eq|by eauto]
  end.


Notation depmach0 :=
  (fun _ => mkmach _ _ _ _).

Notation depmach1 :=
  (fun cf => mkmach _ ((fun rf => _) (lang_tgt.chp cf))
                      ((fun st => _) (lang_tgt.chp cf))
                      ((fun hp => _) (lang_tgt.chp cf))).

Notation depmach2 :=
  (fun cf cf' => mkmach _ ((fun rf rf' => _) (lang_tgt.chp cf) (lang_tgt.chp cf'))
                          ((fun st st' => _) (lang_tgt.chp cf) (lang_tgt.chp cf'))
                          ((fun hp hp' => _) (lang_tgt.chp cf) (lang_tgt.chp cf'))).


Ltac depmcplus0 :=
  instantiate (1 := depmach0);
  eapply tgt_mcplus; first (by s; eauto);
  rr; esplits; s; tmext_simpl; try pmplus_eq; eauto.

Ltac depmcplus1 :=
  instantiate (1 := depmach1);
  eapply tgt_mcplus; first (by s; eauto);
  rr; esplits; s; tmext_simpl; try pmplus_eq; eauto.


Ltac dance_esplits :=
  econs; apply and_with_l; splits; [|apply and_r_without]; splits.

Ltac dance_init :=
  eapply dance_S; intros cf XTR mm MM;
  destruct MM as (c' & RLm & RL); unnw;
  destruct cf, c', mm; destruct XTR as [XTR XTR0];
  s in XTR; s in XTR0; clarify;
  destruct RLm as (PC & RF & ST & SV & HP & HV & FIN); unnw;
  s in PC; s in RF; s in ST; s in SV; s in HP; s in HV; s in FIN; clarify;
  destruct RL as (PC & RF & ST & SV0 & HP & HV0 & FIN0); unnw;
  s in PC; s in RF; s in ST; s in SV0; s in HP; s in HV0; s in FIN0; clarify;
  clear SV HV FIN; rewrite <-?pmplus_assoc in *;
  dance_esplits; rewrite ?tmext_shadow; try setmrf mrf;
  [|by intros STEP ? ? STEP'; hexploit (tgt_det STEP STEP'); vauto; i; des; eauto| |
  clear SV0 HV0 FIN0].



Tactic Notation "frames_dep" integer(i) hyp(real) hyp(xtra) ident(frames) :=
  
  do i (
    depdes frames; s in real;
    match type of xtra with appcontext[vcons ?hd] =>
      let tmp := fresh in rename xtra into tmp;
      assert (xtra: forall cfa, vIn cfa frames -> extra cfa) by auto;
      clear tmp hd
    end);
  depdes frames; s in real;
  hexploit xtra; [by econs 1|];
  match goal with |- extra ?hd -> _ =>
    destruct hd; intro xtra'; destruct xtra' as [xtra' xtra''];
    s in xtra'; s in xtra''; subst
  end;
  match type of xtra with appcontext[vcons ?hd] =>
    let tmp := fresh in rename xtra into tmp;
    assert (xtra: forall cfa, vIn cfa frames -> extra cfa) by auto;
    clear tmp
  end.

Tactic Notation "frames_nodep" hyp(real) ident(frames) :=
  unfold dmach_apply in real; clear dependent frames.


Lemma tgt_stack_transfer: forall rf freest vs
    (ST: forall a,
           a >= rf r_sp <-> (exists v, pmread ({rf r_sp |=> vs} ## freest) a v)),
  forall a,
    a >= rf r_sp + length vs <-> (exists v: word, pmread freest a v).
Proof.
  i; dup ST; specialize (ST0 (rf r_sp)); hdesH ST0.
  specialize (ST1 (le_n _)); des; rrdes ST1; clear RD.

  specialize (ST a); des; split.
  - i; hexploit ST; first omega; i; des.
    apply pmread_plusD in X; des; eauto.
    apply pmcells_readD in X; des; subst; exfalso; omega.
  - i; des; hexploit ST0.
    + econs; eapply pmread_plus; eauto; pmplus_eq'.
    + i; clear -H0 H VLD; induction vs using rev_ind; s; first omega.
      rewrite pmcells_app -pmplus_assoc pmplus_swap in VLD.
      assert (a <> rf r_sp + length vs).
      { eapply pmvalid_cell_cell; rrdes H; subst.
        rewrite ->(pmplus_comm _ h'), ->!(pmplus_swap {_ + _|=>_}) in VLD.
        rewrite ->pmplus_comm, <-pmcells_sng in *; decomp_pmvalid VLD; eauto.
      }
      rewrite app_length; s; hexploit IHvs; last omega. 
      eapply pmvalid_plusD; eauto.
Qed.


Lemma tgt_confpred_pop: forall mrf mrf' db freest hp popped
    (CP: tgt_confpred (mrf, db) {| cpc := Nothing; crf := Just mrf;
                                   cst := freest; chp := hp |})
    (SP: mrf' r_sp = mrf r_sp - length popped)
    (VLD: pmvalid ({mrf' r_sp |=> popped} ## freest)),
  tgt_confpred (mrf', db) {| cpc := Nothing; crf := Just mrf';
                            cst := {mrf' r_sp |=> popped} ## freest; chp := hp |}.
Proof.
  i; rrdes CP; ss; clean.
  rr; s; splits; auto.
  
  split; i; des.
  - destruct (le_or_lt (mrf r_sp) a).
    + 
      apply ST in H0; des.
      econs; eapply pmread_plus; eauto; pmplus_eq'.
    + 
      hexploit (@nth_error_some' (a - mrf' r_sp) _ popped); first omega.
      i; des; econs; eapply pmread_plus; first last; eauto.
      eapply (pmcells_read _ _ popped H1); omega.
  - apply pmread_plusD in H; des.
    + hexploit pmcells_readD; eauto; i; des; omega.
    + hdes; hexploit ST1; [by econs; exact H|omega].
Qed.

Lemma tgt_confpred_push: forall mrf mrf' db freest hp vs
    (CP: tgt_confpred (mrf, db) {| cpc := Nothing; crf := Just mrf;
                                   cst := {mrf r_sp |=> vs} ## freest; chp := hp |})
    (SP: mrf' r_sp = mrf r_sp + length vs),
  tgt_confpred (mrf', db) {| cpc := Nothing; crf := Just mrf';
                             cst := freest; chp := hp |}.
Proof.
  i; rrdes CP; ss; clean.
  rr; s; splits; auto.
  rewrite SP; eauto using tgt_stack_transfer.
Qed.                            

Lemma tgt_confpred_extend: forall rf db rf' q v st hp hp'
    (CP: tgt_confpred (rf, db)
                      {|cpc := Nothing; crf := Just rf; cst := st; chp := hp|})
    (REP: repr q v hp')
    (SP: rf r_sp = rf' r_sp)
    (VLD: pmvalid (hp' ## hp)),
  tgt_confpred (rf', extend db q v)
               {| cpc := Nothing; crf := Just rf'; cst := st; chp := hp' ## hp |}.
Proof.
  i; rrdes CP; clean.
  rr; esplits; eauto; s.
  - rewrite -SP; tmext_simpl; eauto.
  - i; rrdes IN; subst; s; subst; last first.
    + eapply repr_mon; [|by eauto].
      s; i; eapply pmread_plus; eauto; pmplus_eq'.
    + eapply repr_mon; last first; eauto using pmread_plus.
Qed.



(** * Utilities specific to intermediate language. *)


Definition fvars: list var -> list label :=
  fold_right (fun v acc => match v with inl (inl l) => l :: acc | _ => acc end) nil.

Definition tvars: list var -> list termvar :=
  fold_right (fun v acc => match v with inl (inr x) => x :: acc | _ => acc end) nil.

Definition kvars: list var -> list contvar :=
  fold_right (fun v acc => match v with inr k => k :: acc | _ => acc end) nil.


Lemma kvars_in: forall k G
    (IN: In (inr k) G),
  In k (kvars G).
Proof.
  induction G; [by auto|]; i.
  s in IN; des; subst; s; [by auto|].
  destruct a; s; eauto.
Qed.

Lemma kvars_app: forall xs xs',
  kvars (xs ++ xs') = kvars xs ++ kvars xs'.
Proof.
  i; induction xs; auto.
  destruct a; ss; f_equal; eauto.
Qed.

Lemma kvars_purge: forall vars,
  kvars (purge vars) = nil.
Proof.
  induction vars; s; auto.
  destruct a; s; try rewrite IHvars; eauto.
Qed.

Lemma tvars_purge: forall vars,
  tvars (purge vars) = tvars vars.
Proof.
  induction vars; s; auto.
  destruct a; s; try rewrite IHvars; eauto.
Qed.

Lemma fvars_purge: forall vars,
  fvars (purge vars) = fvars vars.
Proof.
  induction vars; s; auto.
  destruct a; s; try rewrite IHvars; eauto.
Qed.

Lemma fvars_inl_inl: forall xs,
  fvars (map (inl<*>inl) xs) = xs.
Proof.
  induction xs; s; f_equal; eauto.
Qed.
