(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Omega Program ClassicalChoice.
Require Import ssr vgtac basis myclassical lang_common lang_spec.
Require Export lang_tgt_syntax.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".



Definition enc: instr -> word :=
  enum.

Lemma enc_inj: forall i j, enc i = enc j -> i = j.
Proof.
  exact enum_inj.
Qed.


Definition regfile := @tmap reg _ word.
Definition stack := @pmap word _ word.
Definition heap := @pmap word _ word.
Structure machine :=
  mkmach {mpc: word; mrf: regfile; mst: stack; mhp: heap}.


(** Operational Semantics *)


Definition add_offset n soff := 
  match soff with
    | o_pos m => m + n
    | o_neg m => n - m
  end.

Inductive eval rf st h : operand -> word -> Prop :=
| ev_imm w : eval rf st h (o_imm w) w
| ev_reg r : eval rf st h (o_reg r) (rf r)
| ev_sind r soff w  
    (RD: pmread st (add_offset (rf r) soff) w) :
  eval rf st h (o_sind r soff) w
| ev_hind r soff w  
    (RD: pmread h (add_offset (rf r) soff) w) :
  eval rf st h (o_hind r soff) w
.
Hint Constructors eval.

Inductive store pc rf : stack -> heap -> operand -> word -> machine -> Prop :=
| sto_reg st h r w :
  store pc rf st h (o_reg r) w (mkmach (1+pc) [w//r]>>rf st h)
| sto_sind st h r soff v w :
  store pc rf ({add_offset (rf r) soff |-> v} ## st) h (o_sind r soff) w
    (mkmach (1+pc) rf ({add_offset (rf r) soff |-> w} ## st) h)
| sto_hind st h r soff v w :
  store pc rf st ({add_offset (rf r) soff |-> v} ## h) (o_hind r soff) w
    (mkmach (1+pc) rf st ({add_offset (rf r) soff |-> w} ## h))
.
Hint Constructors store.


(** We assume a deterministic (but otherwise arbitrary) memory allocator. *)

Definition valid_allocator (alloc: heap -> nat -> nat) :=
  forall h n (FIN: pmfinite h) (VLD: pmvalid h),
    pmvalid ({alloc h n |=> repeat 0 n} ## h).

Definition raw_heap := list (word * word).

Axiom raw_alloc: raw_heap -> word -> word.
Axiom raw_alloc_allocator: forall rh i n
    (LT: i < n),
  lookup (i + raw_alloc rh n) rh = None.

Notation cid := constructive_indefinite_description.

(** Lift [raw_alloc] to the more abstract notion of heap used by the
    operational semantics. *)
Definition alloc (h: heap) (n: word) : word :=
  match LEM (pmfinite h), LEM (pmvalid h) with
    | left FIN, left VLD =>
      let findom := proj1_sig (cid _ FIN) in
      let hf := proj1_sig (cid _ (pmvalid_some VLD)) in
      let f l := match hf l with Some v => [(l, v)] | _ => [] end in
      raw_alloc (flatten (map f findom)) n
    | _, _ => 42 
  end.

Lemma alloc_valid:
  valid_allocator alloc.
Proof.
  unfold valid_allocator, alloc; i; des_ifs.
  destruct (cid _ p); destruct (cid _ (pmvalid_some p0)).
  subst.

  set (rh := flatten _).

  assert (forall l v, pmread (Some x0) l v <-> lookup l rh = Some v).
  {
    split; i.
    - dup H; apply i in H; apply in_split_first in H; des.
      apply pmread_some in H0.
      unfold rh; subst; s; clear -H0 H1; induction l1.
      { s; rewrite H0 -cons_app; des_ifs. }
      exploit IHl1; [by ss; eauto|]; i; s.
      rewrite -x1; s; des_ifs.
      rewrite -cons_app list_assoc_cons_neq; auto.
      ii; subst; ss; eauto.
    - apply list_assoc_someD in H; apply flatten_inD in H; des.
      apply in_map_iff in INb; des.
      des_ifs; s in INa; des; vauto.
      apply pmread_some; auto.
  }

  apply pmvalid_plus_read; auto; []; i; apply H in RDa0.
  apply pmcells_readD in RDa; des.
  assert (X := @raw_alloc_allocator rh (x1 - raw_alloc rh n) n).
  hexploit X; [rewrite repeat_length in LE0; omega|].
  replace (x1 - _ + _) with x1; [|omega]; i; vauto.
Qed.



Inductive step_ok : event -> machine -> machine -> Prop := 
| Sjmp: forall pc rf st h addr pc'
    (INS: pmread h pc (enc (i_jmp addr)))
    (EVAL: eval rf st h addr pc'),
  step_ok etau (mkmach pc rf st h) (mkmach pc' rf st h)
| Sjz: forall pc rf st h r addr
    (INS: pmread h pc (enc (i_jnz r addr)))
    (EVALV: rf r = 0),
  step_ok etau (mkmach pc rf st h) (mkmach (1+pc) rf st h)
| Sjnz: forall pc rf st h r n addr pc'
    (INS: pmread h pc (enc (i_jnz r addr)))
    (EVALV: rf r = S n)
    (EVALA: eval rf st h addr pc'),
  step_ok etau (mkmach pc rf st h) (mkmach pc' rf st h)
| Sld: forall pc rf st h r src w
    (INS: pmread h pc (enc (i_ld r src)))
    (EVAL: eval rf st h src w),
  step_ok etau (mkmach pc rf st h) (mkmach (1+pc) [w//r]>>rf st h) 
| Ssto: forall pc rf st h tgt src m
    (INS: pmread h pc (enc (i_sto tgt src)))
    (UPD: store pc rf st h tgt (rf src) m),
  step_ok etau (mkmach pc rf st h) m
| Slpc: forall pc rf st h r
    (INS: pmread h pc (enc (i_lpc r))),
  step_ok etau (mkmach pc rf st h) (mkmach (1+pc) [pc//r]>>rf st h) 
| Sopr: forall pc rf st h op tgt arga argb va vb m
    (INS: pmread h pc (enc (i_opr op tgt arga argb)))
    (EVALa: eval rf st h arga va)
    (EVALb: eval rf st h argb vb)
    (UPD: store pc rf st h (o_reg tgt) (bosem op va vb) m),
  step_ok etau (mkmach pc rf st h) m
| Salc: forall pc rf st h r rsize loc
    (INS: pmread h pc (enc (i_alc r rsize)))
    (ALC: loc = alloc h (rf rsize)),
  step_ok etau (mkmach pc rf st h)
    (mkmach (1+pc) [loc//r]>>rf st ({loc |=> repeat 0 (rf rsize)} ## h))
| Sinp: forall pc rf st h r n
    (INS: pmread h pc (enc (i_inp r))),
  step_ok (ein n) (mkmach pc rf st h) (mkmach (1+pc) [n//r]>>rf st h) 
| Sout: forall pc rf st h r
    (INS: pmread h pc (enc (i_out r))),
  step_ok (eout (rf r)) (mkmach pc rf st h) (mkmach (1+pc) rf st h)
.

Definition mkerr m := mkmach m.(mpc) m.(mrf) m.(mst) None.

Inductive tgt_step : event -> machine -> machine -> Prop :=
| Sok:  forall evt m m'
          (NZ: ~ m.(mpc) = 0)
          (ST: step_ok evt m m'),
        tgt_step evt m m'
| Serr: forall m
          (NZ: m.(mpc) <> 0)
          (ST: forall evt m', ~ step_ok evt m m'),
        tgt_step etau m (mkerr m)
.
Hint Constructors tgt_step.

Definition tgt_halted m := m.(mpc) = 0. 


Definition o_zero := o_pos 0.
Definition o_one  := o_pos 1.
Definition o_mone := o_neg 1.



(** Loading and Linking *)


Structure group :=
  { g_cptrs: list (label * word)
  ; g_data: list word
  }.

Definition module :=
  list group.

Definition link: module -> module -> module :=
  @app _.



Fixpoint load (m: module) laddr imports : heap * heap * list (label * word) :=
  match m with
    | nil => (pmemp, pmemp, nil)
    | g :: m' =>
      let ni := length imports in
      let ne := length (g_cptrs g) in
      let nd := length (g_data g) in

      let cptrs := map (fun p => laddr + ni + ne + ne + snd p) (g_cptrs g) in
      let exports := swap (enumerate (laddr + ni + ne) (map fst (g_cptrs g))) in

      let gh := {laddr + ni + ne |=> cptrs} in
      let lh := {laddr |=> imports ++ map snd exports} ## {laddr + ni + ne + ne |=> g_data g} in

      let '(gh', lh', exports') := load m' (laddr + ni + ne + ne + nd) (imports ++ map snd exports) in
      (gh ## gh', lh ## lh', exports ++ exports')
  end.
                          
Definition tgt_load (m: module) : option machine :=
  let '(gh, lh, exported) := load m 0 nil in
  match lookup MAIN exported with
    | Some w =>
      Some {| mpc := match gh with
                       | Some hf => match hf w with Some a => a | _ => 42 end
                       | _ => 42
                     end;
              mrf := [w//r_clo]>>(fun _ => 0);
              mst := Some (fun _ => Some 0);
              mhp := gh ## lh |}
    | _ =>
      None
  end.




Structure config :=
  mkconf {cpc: eoption word; crf: eoption regfile; cst: stack; chp: heap}.

Definition realizes (c: config) (m: machine) : Prop :=
  <<PC: c.(cpc) = Just m.(mpc)>> /\
  <<RF: c.(crf) = Just m.(mrf)>> /\
  <<ST: c.(cst) = m.(mst)>> /\ <<SV: pmvalid m.(mst)>> /\
  <<HP: c.(chp) = m.(mhp)>> /\ <<HV: pmvalid m.(mhp)>> /\ <<FIN: pmfinite m.(mhp)>>.


(** Internally the module stores a map from labels (exports) to the
    beginning of their code.  Loading produces a map from labels to
    the closure value!
*)


Definition tgt_wfmod (m: module) exp :=
  <<DOM: exp = flatten (map (map fst <*> g_cptrs) m)>> /\
  <<BND: forall g (INg: In g m)
             l w (INl: In (l, w) (g_cptrs g)),
           w < length (g_data g)>>.






































































Section LangSpecImpl.

  Definition tgt_val := word.
  Definition tgt_cont := word.
  Definition tgt_conf := config.
  Definition tgt_mach := machine.
  Definition tgt_module := module.
  Definition tgt_anchor := word.

  Instance Enumerable_tgt_val: Enumerable tgt_val :=
    {| enum := enum; enum_inj := enum_inj |}.

  Definition tgt_vload (m: module) laddr (imports: list (label * word)) lbl v :=
    
    lookup lbl (snd (load m laddr (map snd imports))) = Some v.

  Definition tgt_cload (m: module) laddr (imports: list (label * word)) (cg cl: config) :=
    
    exists hp_imports rf,
      let '(gh, lh, exported) := load m laddr (map snd imports) in
      <<GPC: cg.(cpc) = Nothing>> /\
      <<GRF: cg.(crf) = Just rf>> /\
      <<GST: rf r_sp = 0 /\ forall a, pmread cg.(cst) a 0>> /\
      <<GHP: cg.(chp) = hp_imports ## gh>> /\
      <<GDJ: pmvalid cg.(chp)>> /\
      <<GDOM: forall loc, In loc (map snd imports) <-> exists v, pmread hp_imports loc v>> /\
      <<LPC: cl.(cpc) = Nothing>> /\
      <<LRF: cl.(crf) = Nothing>> /\
      <<LST: cl.(cst) = pmemp>> /\
      <<LHP: cl.(chp) = lh>> /\
      <<LDJ: pmvalid (hp_imports ## cl.(chp))>>.

  Definition tgt_cplus (c c': tgt_conf) : tgt_conf :=
    match c, c' with {| cpc := cpc; crf := crf; cst := cst; chp := chp |},
                     {| cpc := cpc'; crf := crf'; cst := cst'; chp := chp' |} =>
      {| cpc := eoplus cpc cpc'
       ; crf := eoplus crf crf'
       ; cst := cst ## cst'
       ; chp := chp ## chp'
      |}
    end.

  Definition tgt_cemp :=
    {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := pmemp |}.

  Definition tgt_core c := exists pc,
    c.(cpc) = Just pc /\ c.(crf) = Nothing /\ c.(cst) = pmemp /\ c.(chp) = pmemp.
  Definition tgt_extra c := c.(cpc) = Nothing /\ c.(crf) = Nothing.
  Definition tgt_real := realizes.

  Lemma store_st: forall pc rf st hp tgt src m
      (UPD: store pc rf st hp tgt src m)
      (VLD: pmvalid (mst m)),
    pmvalid st.
  Proof.
    i; inv UPD; eauto using pmvalid_plus_cell.
  Qed.

  Lemma store_hp: forall pc rf st hp tgt src m
      (UPD: store pc rf st hp tgt src m)
      (FIN: pmfinite (mhp m))
      (VLD: pmvalid (mhp m)),
    pmfinite hp /\ pmvalid hp.
  Proof.
    i; inv UPD; split; eauto using pmvalid_plus_cell.
    ss; split_pmfinite FIN VLD; eauto.
  Qed.

  Definition tgt_lsi: LangSpec.
    refine
      {| val := tgt_val
       ; cont := tgt_cont
       ; conf := tgt_conf
       ; mach := tgt_mach
       ; lang_spec.module := tgt_module
       ; anchor := tgt_anchor
       ; vload := tgt_vload
       ; cload := tgt_cload
       ; cplus := tgt_cplus
       ; cemp := tgt_cemp
       ; halted := tgt_halted
       ; core := tgt_core
       ; extra := tgt_extra
       ; lang_spec.step := tgt_step
       ; real := tgt_real
      |}.
    - split.
      + i; destruct m; unfold tgt_halted in *; ss.
        inversion FH; vauto.
      + i; r; s; apply NNPP; intro NV.
        destruct (classic (exists evt m', step_ok evt m m')); des; eauto.
    - destruct a, b; unfold tgt_cplus; s.
      f_equal; eauto using pmplus_comm, eoplus_comm.
    - i; destruct a, b, c; s; f_equal; eauto using pmplus_assoc, eoplus_assoc.
    - i; destruct a; s; destruct cpc0, crf0; rewrite !pmplus_emp; auto.
    - i; apply RL; i; rrdes FH0; destruct c; ss; clarify.
      inv ST; [|by rrdes HV; vauto].
      inv ST0; ss; try (by eapply (FH (mkconf _ _ _ _)); econs; esplits; ss; eauto).
      + hexploit store_st; eauto; i; hexploit store_hp; eauto; i; des.
        eapply (FH (mkconf _ _ _ _)); econs; esplits; ss; eauto.
      + hexploit store_st; eauto; i; hexploit store_hp; eauto; i; des.
        eapply (FH (mkconf _ _ _ _)); econs; esplits; ss; eauto.
      + split_pmfinite FIN HV.
        eapply (FH (mkconf _ _ _ _)); rr; esplits; ss; eauto.
    - i; apply NNPP; ii; inversion ST; vauto; inversion ST0; vauto
      ; apply NE; i; rrdes FH; ss; clarify
      ; eapply (E (mkconf _ _ _ _)); rr; esplits; ss; eauto.
    - i; invc COa COb RLa RLb; des; subst.
      destruct m, coa, ca, cob, cb; ss; noemp; clarify.
      destruct cpc1, cpc3; noemp; subst; vauto.
    - ss; clarify.
    - i; destruct c, c'; rrdes XTR XTR'; ss; subst; rr; eauto.
    - set (dummy := {| mpc := 0; mrf := fun _ => 0; mst := pmemp; mhp := pmemp |}).
      i; destruct c.
      destruct cpc0; [|M|M]
      ; Mdo (by exists dummy; i; destruct cf; destruct cpc0; inv XTR; rrdes RL; vauto).
      destruct crf0; [|M|M]
      ; Mdo (by exists dummy; i; destruct cf; destruct crf0; inv XTR; rrdes RL; vauto).
      econs; instantiate (1 := {| mpc := _; mrf := _; mst := _; mhp := _ |}). 
      destruct cf; ss; clarify; eauto.
      i; inv XTR; rrdes RL; rewrite <-HP, <-ST in *; ss.
      split_pmfinite FIN HV; split_pmvalid SV; rr; esplits; ss; eauto.
      
      
      
      
      
    - i; destruct m; unfold tgt_halted in *; ss; des.
      destruct cf; rrdes XTR; ss; clarify.
      destruct mm; rrdes RL; unfold tgt_cplus in *; destruct c'; ss; clarify.
      destruct cpc0, crf0; rrdes RLm; ss; clarify.
  Defined.

  Definition tgt_link := link.

End LangSpecImpl.

Hint Unfold tgt_core.





Lemma swap_fst: forall X Y (l: list (X * Y)),
  map fst (swap l) = map snd l.
Proof.
  induction l; vauto; destruct a; s; f_equal; eauto.
Qed.

Lemma swap_snd: forall X Y (l: list (X * Y)),
  map snd (swap l) = map fst l.
Proof.
  induction l; vauto; destruct a; s; f_equal; eauto.
Qed.

Lemma enumerate_snd: forall X (l: list X) n,
  map snd (enumerate n l) = l.
Proof.
  induction l; vauto; i; s; f_equal; eauto.
Qed.                            


Lemma load_exports: forall m oenv gh lh exports laddr imports
    (WF: tgt_wfmod m oenv)
    (E: (gh, lh, exports) = load m laddr imports),
  oenv = map fst exports.
Proof.
  induction m; i; ss; clarify; rrdes WF; vauto.
  destruct (load _ _ _) as [[gh' lh'] exports'] eqn:E'.
  des; subst; clarify; s.
  rewrite map_app swap_fst enumerate_snd; f_equal.
  eapply IHm; eauto; split; ss; eauto.
Qed.

Lemma load_valid_aux: forall n junk m laddr imports gh lh exports
    (E: load m laddr imports = (gh, lh, exports))
    (Eladdr: n + length junk = laddr),
  pmvalid ({n |=> junk} ## gh ## lh).
Proof.
  i; ginduction m; i; s; [by inv E; noemp|].
  ss; destruct (load m _ _) as [[gh' lh'] exports'] eqn:E'; s; clarify.
  rewrite -!pmplus_assoc 2!(pmplus_swap gh') -(pmplus_swap {n + _ |=> _}).
  rewrite pmplus_assoc -pmcells_app.
  rewrite -plus_assoc pmplus_assoc -plus_assoc.
  rewrite -pmcells_app'; last first.
  { rewrite !app_length swap_snd map_length enumerate_length map_length; auto. }
  rewrite -plus_assoc pmplus_assoc -pmcells_app'; last first.
  { rewrite !app_length swap_snd !map_length enumerate_length map_length; auto. }
  eapply IHm; eauto.
  rewrite swap_snd !app_length !plus_assoc !map_length enumerate_length map_length; auto.
Qed.

Lemma load_valid: forall m laddr imports gh lh exported
    (E: load m laddr imports = (gh, lh, exported)),
  pmvalid (gh ## lh).
Proof.
  i; hexploit (load_valid_aux laddr nil); eauto; []; i.
  rewrite pmcells_nil in X; noemp; auto.
Qed.  

Lemma load_gfinite: forall m laddr imports gh lh exported
    (E: load m laddr imports = (gh, lh, exported)),
  pmfinite gh.
Proof.
  induction m; i; s; [by inv E; noemp|].
  ss; destruct (load m _ _) eqn:E'; destruct p as [gh' lh']; s; clarify; eauto.
Qed.

Lemma load_lfinite: forall m laddr imports gh lh exported
    (E: load m laddr imports = (gh, lh, exported)),
  pmfinite lh.
Proof.
  induction m; i; s; [by inv E; noemp|].
  ss; destruct (load m _ _) eqn:E'; destruct p as [gh' lh']; s; clarify; eauto.
Qed.

Lemma load_in: forall m laddr imports l v gh lh exports
    (E: load m laddr imports = (gh, lh, exports))
    (IN: In (l, v) exports),
  exists w, pmread gh v w.
Proof.
  i; hexploit load_valid; eauto; []; i.
  ginduction m; ss; i; clarify; vauto.
  destruct (load _ _ _) as [[gh' lh'] exports'] eqn:E'; clarify.
  apply in_app_or in IN; des.
  - dup IN; eapply swap_inD, enumerate_inD in IN; des.
    apply in_map_iff in IN1.
    apply swap_inD, enumerate_inD in IN0.
    des; destruct x; subst l.
    apply pmvalid_plusD in X; des; clear X0.

    assert (YY: exists w, nth_error (map
        (fun p => laddr + length imports + length (g_cptrs a) + length (g_cptrs a) + snd p)
        (g_cptrs a)) (v - (laddr + length imports + length (g_cptrs a))) = Some w).
    { eapply nth_error_some'; rewrite ->map_length in *; omega. }
    
    des; econs; eapply pmread_plus; [| |by eauto]; [|by eauto].
    apply pmvalid_plusD in X; des; clear X0; s in IN0.
    des; eapply pmcells_read in YY; eauto; omega.
  - dup X; rewrite <-!pmplus_assoc, 2(pmplus_swap gh') in X; decomp_pmvalid X.
    hexploit IHm; eauto; []; i; des.
    econs; eapply pmread_plus; eauto.
    rewrite ->pmplus_comm; eauto.
Qed.

Lemma load_out: forall m laddr imports gh lh exports v w
    (E: load m laddr imports = (gh, lh, exports))
    (IN: pmread gh v w),
  exists l, In (l, v) exports.
Proof.
  clear.
  induction m; ss; i; clarify; [by apply pmread_empD in IN; vauto|].
  destruct (load _ _ _) as [[gh' lh'] exports'] eqn:E'; clarify.
  apply pmread_plusD in IN; des.
  - apply pmcells_readD in IN; des; clear NTH.
    rewrite map_length in LE0.
    pattern (length (g_cptrs a)) at 2 in LE0.
    replace (length (g_cptrs a)) with (length (map fst (g_cptrs a))) in LE0
    ; [|by rewrite map_length].
    hexploit enumerate_in; eauto; []; i; des.
    apply swap_in in X; eauto using in_or_app.
  - eapply IHm in IN; [|by eauto]; i; des; eauto using in_or_app.
Qed.





















































































Lemma tgt_cload_finite: forall m laddr imports cg cl
    (LD: tgt_cload m laddr imports cg cl),
  pmfinite cg.(chp) /\ pmfinite cl.(chp).
Proof.
  i; unfold tgt_cload in LD; destruct (load _ _ _) as [[gh lh]] eqn:E; des.
  rewrite GHP LHP.
  split; [apply pmfinite_plus|]; eauto using load_gfinite, load_lfinite.
  exists (map snd imports); i; apply GDOM; eauto.
Qed.




Lemma eval_uniq: forall rf st hp op wa wb
    (EVa: eval rf st hp op wa)
    (EVb: eval rf st hp op wb),
  wa = wb.
Proof.
  i; inv EVa; inv EVb; eauto using (@pmread_uniq nat).
Qed.

Lemma store_uniq: forall pc rf st hp tgt src ma mb
    (STa: store pc rf st hp tgt src ma)
    (STb: store pc rf st hp tgt src mb)
    (VLDhp: pmvalid hp)
    (VLDst: pmvalid st),
  ma = mb.
Proof.
  i; inv STa; inv STb; try rename H1 into H0
  ; apply eq_sym, pmcell_uniq in H0; des; vauto.
Qed.





Theorem tgt_det: forall m evta evtb ma mb
    (STa: tgt_step evta m ma)
    (STb: tgt_step evtb m mb)
    (VLDhp: pmvalid (mhp m))
    (VLDst: pmvalid (mst m))
    (EVT: forall n, evta <> ein n),
  evta = evtb /\ ma = mb.
Proof.
  i; invc STa; invc STb
  ; [|by (exfalso; eapply ST0; eauto)|by (exfalso; eapply ST; eauto)].
  invc ST; invc ST0; hexploit (pmread_uniq INS INS0); intro ENC
  ; apply enc_inj in ENC; vauto.
  - hexploit (eval_uniq EVAL EVAL0); i; inst; subst; eauto.
  - rewrite ->EVALV in *; vauto.
  - rewrite ->EVALV in *; vauto.
  - rewrite ->EVALV in *; vauto.
    hexploit (eval_uniq EVALA EVALA0); i; inst; subst; eauto.
  - hexploit (eval_uniq EVAL EVAL0); i; inst; subst; eauto.
  - hexploit (store_uniq UPD UPD0); i; inst; subst; eauto.
  - hexploit (eval_uniq EVALa EVALa0); i; inst; subst; eauto.
    hexploit (eval_uniq EVALb EVALb0); i; inst; subst; eauto.
    hexploit (store_uniq UPD UPD0); i; inst; subst; eauto.
  - exfalso; eapply EVT; auto.
Qed.

Definition instr_wdet instr : Prop :=
  match instr with
    | i_inp _ | i_alc _ _ => False
    | _ => True
  end.

Lemma mkerr_error: forall m,
  error tgt_lsi (mkerr m).
Proof.
  i; destruct m; ii.
  rrdes FH; ss; rrdes HV; vauto.
Qed.

               
Lemma eval_frame: forall rf st stf hp hpf w w'
    (EVAL: eval rf st hp w w')
    (VLDs: pmvalid (st ## stf))
    (VLDh: pmvalid (hp ## hpf)),
  eval rf (st ## stf) (hp ## hpf) w w'.
Proof.
  i; inv EVAL; [econs 3|econs 4]; eapply pmread_plus; eauto.
Qed.

Definition mplus_sh m st hp :=
  {| mpc := mpc m; mrf := mrf m; mst := mst m ## st; mhp := mhp m ## hp |}.

               
Lemma store_frame: forall pc rf st stf hp hpf op w m
    (STO: store pc rf st hp op w m)
    (VLDs: pmvalid (st ## stf))
    (VLDh: pmvalid (hp ## hpf)),
  store pc rf (st ## stf) (hp ## hpf) op w (mplus_sh m stf hpf) /\
  pmvalid (mst m ## stf) /\ pmvalid (mhp m ## hpf) /\
  (pmfinite (hp ## hpf) -> pmfinite (mhp m ## hpf)).
Proof.
  i; inv STO; unfold mplus_sh; s; try rewrite <-!pmplus_assoc
  ; (splits; [by econs| | |]); eauto.
  - rewrite <-pmplus_assoc in VLDs; eapply pmvalid_plus_cell; eauto.
  - rewrite <-pmplus_assoc in VLDh; eapply pmvalid_plus_cell; eauto.
  - intro FIN; rewrite <-pmplus_assoc in VLDh; split_pmfinite FIN VLDh.
    repeat apply pmfinite_plus; eauto using pmfinite_cell.
Qed.

Lemma step_ok_frame: forall evt m inst m' cf mf
    (PC: pmread (mhp m) (mpc m) (enc inst))
    (DI: instr_wdet inst)
    (ST: step_ok evt m m')
    (PL: mcplus tgt_lsi m cf mf),
  step_ok evt mf (mplus_sh m' (cst cf) (chp cf)).
Proof.
  i; rrdes PL; destruct mf, cf, c'; inv ST
  ; unfold mplus_sh; rrdes RL RLm; ss; clarify
  ; try (by hexploit (pmread_uniq PC INS); intro E; apply enc_inj in E; clarify)
  ; eapply pmread_plus in INS; eauto
  ; destruct cpc0, crf0; ss; clarify
  ; eauto using Sjz, Slpc, Sout.
  - eapply eval_frame in EVAL; eauto using Sjmp.
  - eapply eval_frame in EVALA; eauto using Sjnz.
  - eapply eval_frame in EVAL; eauto using Sld.
  - eapply store_frame in UPD; des; eauto using Ssto.
  - eapply eval_frame in EVALa; eauto.
    eapply eval_frame in EVALb; eauto.
    eapply store_frame in UPD; des; eauto using Sopr.
Qed.

Lemma tgt_wdet: forall m inst
    (PC: pmread (mhp m) (mpc m) (enc inst))
    (DI: instr_wdet inst),
  @wdet tgt_lsi m.
Proof.
  ii; invc STa; [|by exfalso; eapply mkerr_error; eauto].
  invc STb; [|by exfalso; eapply ST0, step_ok_frame; eauto].
  rrdes PL; destruct inst; vauto; clear DI NZ
  ; (invc ST
    ; hexploit (pmread_uniq PC INS); intro E; apply enc_inj in E; clarify
    ; destruct cf; rrdes RLm; rrdes RL0; ss; clarify; clear INS
    ; rrdes RL; destruct ca', c'; ss; clarify
    ; destruct crf0, cpc0; vauto; ss; clarify
    ; inv ST0; ss; clarify; eapply pmread_plus in PC; eauto
    ; hexploit (pmread_uniq PC INS); intro E; apply enc_inj in E; clarify
    ; try (by omega)
    ; try (by split; auto; rr; splits; auto)).
  - eapply eval_frame in EVAL; eauto.
    hexploit (eval_uniq EVAL EVAL0); intro E; instantiate; subst.
    split; auto; rr; splits; auto.
  - eapply eval_frame in EVALA; eauto.
    hexploit (eval_uniq EVALA EVALA0); intro E; instantiate; subst.
    split; auto; rr; splits; auto.
  - eapply eval_frame in EVAL; eauto.
    hexploit (eval_uniq EVAL EVAL0); intro E; instantiate; subst.
    split; auto; rr; splits; auto.
  - eapply store_frame in UPD; eauto; des.
    hexploit (store_uniq UPD UPD0); eauto; intro E; instantiate; subst.
    split; auto; rr; splits; auto.
  - eapply eval_frame in EVALa; eauto.
    hexploit (eval_uniq EVALa EVALa0); intro E; instantiate; subst.
    eapply eval_frame in EVALb; eauto.
    hexploit (eval_uniq EVALb EVALb0); intro E; instantiate; subst.
    eapply store_frame in UPD; eauto; des.
    hexploit (store_uniq UPD UPD0); eauto; intro E; instantiate; subst.
    split; auto; rr; splits; auto.
Qed.

Lemma tgt_wfmod_nil: tgt_wfmod [Build_group nil nil] nil.
Proof.
  rr; splits; vauto; i; ss; des; vauto.
Qed.
