(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_src_lsi lang_tgt.
Require Import model_common model model_properties.
Require Import gw_common gw zwickel.
Require Import ssr util ts.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".

Opaque pmplus pmcells.          



Lemma tr_mod_aux_labels: forall a b c m,
  map fst m = map fst (g_cptrs (tr_mod_aux a b c m)).
Proof.
  i; ginduction m; vauto; i; s; destruct a; s; erewrite IHm; eauto.
Qed.


Lemma list_assoc_index: forall X {M: EqDec X} Y {N: EqDec Y} eqdec x y (l: list (X * Y))
    (IN: list_assoc eqdec x l = Some y),
  exists i,
    <<FST: list_index x (map fst l) = Some i>> /\
    <<ALL: list_index (x, y) l = Some i>>.
Proof.
  i; ginduction l; vauto.
  i; destruct a; ss; des_ifs; eauto.
  hexploit IHl; eauto; []; i; des.
  rewrite FST ALL; s; eauto.
Qed.                            


Lemma eq_minus: forall x y z,
  x = z + y -> x - y = z.
Proof.
  i; omega.
Qed.

Lemma ge_minus: forall x y z,
  x >= z + y -> x - y >= z.
Proof.
  i; omega.
Qed.


Lemma nth_error_app': forall X (la lb: list X) n x
    (IN: nth_error lb (n - length la) = Some x)
    (LE: n >= length la),
  nth_error (la ++ lb) n = Some x.
Proof.
  induction la; i; destruct n; vauto.
  - exfalso; ss; omega.
  - ss; apply IHla; eauto with arith.
Qed.                            

Lemma nth_error_cons: forall X (l: list X) n a b
    (IN: nth_error l (n - 1) = Some b)
    (POS: n > 0),
  nth_error (a::l) n = Some b.
Proof.
  i; rewrite cons_app; eauto using nth_error_app'.
Qed.


Ltac readpc_aux pc E :=
  eapply pmcells_read; [|match goal with |- ?lhs = ?x + pc =>
                           instantiate (1 := lhs - pc); omega
                       
                         end];
  repeat (
    simpl map; rewrite ?map_app; (apply nth_error_cons || apply nth_error_app');
    last (unfold tgt_val; rewrite ?map_length; clear; repeat apply ge_minus; omega)
  );
  match goal with |- nth_error ?l ?n = _ =>
    let H := fresh in
    assert (H: n = 0) by (
      unfold tgt_val; rewrite ?map_length;
      repeat (match goal with |- context[length ?l] => generalize (length l) end);
      clear; i;
      repeat (match goal with
                | |- ?a - ?b = _ => apply eq_minus
                | |- _ <= ?a - ?b => apply ge_minus
                | |- _ => omega
              end)
    );
    rewrite H
  end;
  simpl nth_error; eauto.

Ltac readpc' pc E :=
  eapply pmread_plus; last (by eauto);
  last (by rhs rewrite ->E, <-!pmplus_assoc,  <-!(pmplus_swap {pc |=> _}));
  readpc_aux pc E.

Ltac readpc EXT E pc :=
  eapply pmread_plus; first (eapply EXT); eauto; [|by pmplus_eq'];
  rewrite E;
  eapply pmread_plus; [|by eauto|by rewrite -E; eauto];
  readpc_aux pc E.



(** * Local world. *)

Program Definition mylwsts: transys :=
  
  {| state := stack * heap
   ; ext    s' s := <<HP: forall l v (IN: pmread (snd s) l v), pmread (snd s') l v>>
   ; extpub s' s := <<HP: forall l v (IN: pmread (snd s) l v), pmread (snd s') l v>>
                 /\ <<ST: fst s' = fst s>>
  |}.
Next Obligation.
  esplits; ss; [auto|congruence].
Qed.

Program Definition mylwconf: method_conf tgt_lsi src_lsi (tsprod (gwsts ts_gw) mylwsts) :=
  {| confrel g s ca cb :=
       <<Ca: ca = {| cpc := Nothing; crf := Nothing
                   ; cst := fst (snd s); chp := snd (snd s) |}>> /\
       <<Cb: cb = (pmemp, Nothing, Nothing)>>
  |}.

Program Definition mylwname: method_name tgt_lsi src_lsi (tsprod (gwsts ts_gw) mylwsts) :=
  {| supp := fun _ => False
   ; name := fun _ _ _ _ _ => False
  |}.

Program Definition mylw: world_loca ts_gw :=
  {| lwwfo := gwfo
   ; lwsts := mylwsts; lwconf := mylwconf; lwname := mylwname |}.
Next Obligation.
  unfold tgt_extra, src_extra; unnw; subst; ss; eauto.
Qed.


Notation ts_env := (fun s => snd (fst (fst s))).


Lemma mylw_stable: stable mylw.
Proof.
  ii; rrdes CR; rr; esplits; s; eauto.
Qed.

Lemma mylw_confrel_shape: forall s r ca cb
    (CR: confrel mylw s r ca cb),
  exists lsta lhpa,
    ca = {| cpc := Nothing; crf := Nothing; cst := lsta; chp := lhpa |} /\
    cb = (pmemp, Nothing, Nothing).
Proof.
  i; destruct ca, cb; rrdes CR; ss; subst; eauto.
Qed.



(** ** Program similarity. *)

Notation proj_env s := (snd (fst (fst s))).

Definition esim' La Lb (w: world La Lb) i ka kb g si s ty ea eb : Prop :=
  forall csa csb (CR: confrel w g s csa csb),
    esim w i ka kb g si s false ty (ea # csa) (eb # csb).


Definition ts_psim_aux i F G ty ea eb : Prop :=
  forall

    tsub g si s
    (LCT: lc_tsub tsub)
    (GK: gknow (wlift mylw) g)
    (EXT: (wlift mylw).(ext) s si)

    
    senva
    (RSV: forall l ty' (IN: lookup l F = Some ty'), exists va vb,
            <<INa: lookup l senva = Some va>> /\
            <<INb: lookup l (proj_env s) = Some vb>> /\
            <<SIM: vclos (wlift mylw) s (g s) (tsubst_name tsub ty') va vb>>)
    (SORD: map fst F = map fst senva)

    
    denva denvb
    (RDV: forall x ty' (IN: lookup x G = Some ty'), exists va,
            <<INa: lookup x denva = Some va>> /\
            <<SIM: vclos (wlift mylw) s (g s) (tsubst_name tsub ty') va (denvb x)>>)
    (ENV: reprlist (ts_lhp s) (ts_rf s r_env) (map snd denva))
    (DORD: map fst G = map fst denva)
    (CL: forall x (IN: In x (map fst G)), forall y, ~ fv_exp (denvb x) y)
    (WCL: subst_weakly_closed denvb)

    
    laddr n
    (NZ: laddr <> 0)
    (Eladdr: exists h, ts_lhp s = {laddr |=> ea n} ## {laddr - n |=> map snd senva} ## h)

    
    j kia kib ka kb ty'
    (Ka: ka = length (ea n) + laddr)
    (KSIM: forall g' s' va vb
                  (LEG: g <g= g')
                  (GK': gknow (wlift mylw) g')
                  (EXT: ext (wlift mylw) s' s)
                  (ENV: ts_rf s' r_env = ts_rf s r_env)
                  (SP: ts_rf s' r_sp = ts_rf s r_sp + 1)
                  (RET: ts_rf s' r_ret = ts_rf s r_ret)
                  (LST: ts_lst s' = {ts_rf s' r_sp - 1 |-> va} ## ts_lst s)
                  (VR: vclos (wlift mylw) s' (g' s') (tsubst_name tsub ty) va vb),
             esim' (wlift mylw) j kia kib g' si s' ty'
                  {| cpc := Just ka; crf := Nothing
                   ; cst := pmemp; chp := pmemp |}
                  (pmemp, Nothing, Just (app_cont kb vb))),

  esim' (wlift mylw) (wfo_add i j) kia kib g si s ty'
        {| cpc := Just laddr; crf := Nothing; cst := pmemp; chp := pmemp |}
        (pmemp, Nothing, Just (app_cont kb (subst_exp denvb eb))).

Definition ts_psim F G ty ea eb :=
  exists i, ts_psim_aux i F G ty ea eb.



(** * Correctness proof. *)



(** ** Reasoning about [getLabel].
   
    (We don't actually care about [r_aux] but this is simpler than
    quantifying over certain register files in [SIM]).
*)
Lemma pesim_getLabel: forall pc senva o v v' j kia kib g si s l hp eb rf st i ty
    (r: esimtyp (wlift mylw))
    (IN: lookup l senva = Some v)
    (SENV: forall w v (RD: pmread {pc - o |=> map snd senva} w v),
             pmread hp w v)
    (CODE: forall w v (RD: pmread {pc |=> map enc (getLabel l o (map fst senva))} w v),
             pmread hp w v)
    (SIM: pesim r j kia kib g si s false ty
                {| cpc := Just (length (getLabel l o (map fst senva)) + pc)
                 ; crf := Just [v//r_aux]>>[1 + rf r_sp//r_sp]>>rf
                 ; cst := {rf r_sp |-> v} ## st; chp := hp |} eb)
    (LE: wfo_le_n (length (getLabel l o (map fst senva))) j i)
    (NZ: pc <> 0),
  pesim r i kia kib g si s false ty
        {| cpc := Just pc; crf := Just rf; cst := {rf r_sp |-> v'} ## st; chp := hp |} eb.
Proof.
  i; unfold getLabel, push in *.
  s in CODE; s in LE; s in SIM.
  eapply pesim_real; intros ma ? RLa ?.
  destruct ma; destruct_tgt_real RLa; clarify.
  eapply pesim_dance; first (by eauto); clear RLa0 SV HV FIN.

  - dance_init.
    { eapply Sok; [by vauto|eapply Slpc].
      eapply pmread_plus; first (apply CODE); eauto.
      eapply pmread_plus; eauto.
      s; (rhs rewrite pmcells_cons); eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr]; last by eauto.
      - eapply pmread_plus; first (apply CODE); eauto.
        eapply pmread_plus; eauto.
        s; (rhs rewrite 2!pmcells_cons); pmplus_eq'.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - eapply pmread_plus; first (apply CODE); eauto.
        eapply pmread_plus; eauto.
        s; (rhs rewrite !pmcells_cons); eauto; pmplus_eq'.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; first (apply SENV); eauto.
        dup IN; eapply list_assoc_index in IN; des.
        rewrite FST; s.
        apply list_index_split in ALL; des.
        eapply pmread_plus; eauto.
        (rhs rewrite E map_app); s.
        rhs rewrite pmcells_app pmcells_cons map_length LEN plus_comm.
        rhs rewrite pmplus_swap.
        pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr]; last by eauto.
      - eapply pmread_plus; first (apply CODE); eauto.
        eapply pmread_plus; eauto.
        s; (rhs rewrite 4!pmcells_cons); pmplus_eq'.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; first (apply CODE); eauto.
        eapply pmread_plus; eauto.
        s; (rhs rewrite 5!pmcells_cons); pmplus_eq''.
      - unfold stack_top; prep_store; eauto.
    }
    { tmext_simpl; s.
      replace (_ - _) with (mrf0 r_sp); [|omega].
      depmcplus0; eauto using pmvalid_plus_cell.
    }

    eapply dance_O; instantiate (1 := fun _ _ _ _ => depmach0); s; eauto. 

  - i; rrdes XTRb RLb0; clarify.
    destruct eb as [[]]; s in Ec; noemp; clarify. 
    esplits; first 2 last; last first; eauto.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear FIN.
      destruct_tgt_real RLa. eauto.
      rewrite plus_comm tmext_swap; vauto; eauto.
    + s; rr; esplits; eauto.
Qed.


(** ** Reasoning about [getVariable].
   
    (We don't actually care about [r_clo] but this is simpler than
    quantifying over certain register files in [SIM]).
*)
Lemma pesim_getVariable: forall pc denva v v' j kia kib g si s x hp eb rf st i ty
    (r: esimtyp (wlift mylw))
    (IN: lookup x denva = Some v)
    (ENV: reprlist hp (rf r_env) (map snd denva))
    (CODE: forall w v (RD: pmread {pc |=> map enc (getVariable x (map fst denva))} w v),
             pmread hp w v)
    (SIM: pesim r j kia kib g si s false ty
                {| cpc := Just (length (getVariable x (map fst denva)) + pc)
                 ; crf := Just [v//r_clo]>>[1 + rf r_sp//r_sp]>>rf
                 ; cst := {rf r_sp |-> v} ## st; chp := hp |} eb)
    (LE: wfo_le_n (length (getVariable x (map fst denva))) j i)
    (NZ: pc <> 0),
  pesim r i kia kib g si s false ty
        {| cpc := Just pc; crf := Just rf; cst := {rf r_sp |-> v'} ## st; chp := hp |} eb.
Proof.
  i.

  
  eapply pesim_real; intros ma ? RLa ?.
  destruct ma; destruct_tgt_real RLa; clarify.

  eapply pesim_dance; first (by eauto); clear RLa0 SV HV FIN.
  { dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - eapply pmread_plus; first (apply CODE); eauto.
        eapply pmread_plus; eauto.
        s; (rhs rewrite pmcells_cons); eauto.
      - eauto.
    }
    { depmcplus0. }

    eapply dance_O; instantiate (1 := depmach0); eauto.
  }

  
  revert ENV; generalize (mrf r_env) as myenv.
  pose (m := 0); change denva with (skipn m denva).
  replace (S mpc) with (S mpc + m); [|unfold m; omega].
  revert LE; simpl getVariable; simpl length; rewrite app_length repeat_length; simpl length.
  replace (S (index x (map fst denva) + 3)) with (S (index x (map fst denva) + 3 - m))
  ; [|unfold m; omega].
  assert (MLE: myle m (index x (map fst denva))) by (apply myle_le; auto using le_0_n).
  revert MLE i; generalize m; clear m.
  remember (index _ _) as N; intros ? MLE.
  induction MLE.

  { 
    intros ? ?.
    replace (S _) with 4 in LE; [|omega].
    do 3 (apply wfo_le_n_S' in LE; des; rename LE0 into LE).
    i; rrdes XTRb RLb0; clarify.
    destruct eb as [[]]; s in Ec; noemp; clarify.
    esplits; first 1 last.
    - right; split; eauto.
    - inst_pairs; s; noemp; rr; esplits; eauto.
    - i; frames_nodep RLa cfas; simpl plus in RLa.
      destruct ea'.
      esplits; eauto.
      eapply pesim_dance; first (by eauto); clear RLa FIN.
      + 
        rewrite plus_comm.
        rename mrf into mrf0.

        dance_init.
        { eapply Sok; [i; ss; omega|eapply Sopr]; last by eauto.
          - eapply pmread_plus; first (apply CODE); eauto.
            eapply pmread_plus; eauto.
            s; rewrite map_app app_comm_cons pmcells_app.
            simpl length; rewrite repeat_map repeat_length.
            rhs rewrite ->plus_comm, ->?(pmcells_cons mpc), <-?pmplus_assoc; s.
            unfold push; rewrite ->?pmcells_cons; pmplus_eq'.
          - eauto.
          - eauto.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [i; ss; omega|eapply Sld].
          - eapply pmread_plus; first (apply CODE); eauto.
            eapply pmread_plus; eauto.
            s; rewrite map_app app_comm_cons pmcells_app.
            simpl length; rewrite repeat_map repeat_length.
            rhs rewrite ->plus_comm, ->?(pmcells_cons mpc), <-?pmplus_assoc; s.
            unfold push; rewrite ->?pmcells_cons; pmplus_eq''. 
          - dup IN; apply list_assoc_someD in IN.
            apply (in_map fst) in IN; clean.
            eapply list_index_some in IN; des; rewrite IN in ENV.
            eapply skipn_list_assoc in IN; [|by eauto]; des.
            rewrite IN in ENV; s in ENV; des.
            eapply ev_hind; unfold mrf; tmext_simpl; s.
            eapply pmread_plus; eauto.
        }
        { depmcplus0. }
        tmext_uniq.

        dance_init.
        { eapply Sok; [i; ss; omega|eapply Ssto].
          - eapply pmread_plus; first (apply CODE); eauto.
            eapply pmread_plus; eauto.
            s; rewrite map_app app_comm_cons pmcells_app.
            simpl length; rewrite repeat_map repeat_length.
            rhs rewrite ->plus_comm, ->?(pmcells_cons mpc), <-?pmplus_assoc; s.
            unfold push; rewrite ->?pmcells_cons; pmplus_eq''.
          - unfold stack_top; prep_store; eauto.
        }
        { tmext_simpl; s.
          replace (_ - _) with (mrf0 r_sp); [|omega].
          depmcplus0; s; eauto using pmvalid_plus_cell.
        }

        eapply dance_O; instantiate (1 := fun _ _ => depmach0); eauto.

      + 
        i; rrdes XTRb RLb0; clarify.
        s in Ec; noemp; clarify.
        rewrite plus_comm in SIM.
        esplits; first 2 last; last first.
        * right; split; eauto; repeat (econs; last by eauto).
        * i; frames_nodep RLa cfas.
          destruct ea'.
          clear FIN.            
          destruct_tgt_real RLa.
          esplits; eauto.
          s in SIM; rewrite app_length repeat_length in SIM.
          s in SIM; rewrite -(plus_comm 3) plus_comm in SIM; s in SIM.
          s; rewrite -(plus_comm 1); eauto.
        * rr; esplits; s; eauto.
  }
  { 
    i; rrdes XTRb RLb0; clarify.
    destruct eb as [[]]; s in Ec; noemp; clarify.
    inv LE.                     
    esplits; last first.
    - inst_pairs; s; noemp; rr; esplits; eauto.
    - right; split; eauto.
    - i; frames_nodep RLa cfas.
      destruct ea'.
      esplits; eauto.
      dup MLE; eapply myle_le, skipn_cons in MLE; eauto; []; des.
      des; rewrite MLE in ENV; s in ENV; des.

      eapply pesim_dance; first (by eauto); clear RLa FIN.
      + 
        dance_init.
        { eapply Sok; [by vauto|eapply Sld].
          - eapply pmread_plus; first (apply CODE); eauto.
            eapply pmread_plus; eauto.
            s; set (m := index _ _).
            rewrite map_app app_comm_cons pmcells_app.
            rewrite (@repeat_split _ _ (S n)); [|by apply myle_le; eauto].
            rewrite map_app repeat_map repeat_S.
            rewrite pmcells_cons 2!pmcells_app app_length repeat_length; s.
            rewrite ->?pmcells_cons; pmplus_eq''.
          - eapply ev_hind; tmext_simpl; s.
            eapply pmread_plus; eauto.
        }
        { depmcplus0. }
        eapply dance_O; instantiate (1 := depmach0); s; eauto.
      + i; hexploit IHMLE; eauto.
        * s; revert GLE; rewrite plus_comm; simpl plus; i.
          rewrite NPeano.Nat.sub_succ minus_Sn_m; eauto.
          apply myle_le in MLE0; omega.
        * i; nbdes; esplits; eauto.
          i; eapply H; eauto.
          s; s in RLa; rewrite plus_comm; s; rewrite plus_comm.
          rewrite tmext_shadow in RLa; eauto.
  }
Qed.


Ltac ts_prepare' :=
  intro HR;                     
  match type of HR with confrel _ _ _ ?csa _ =>
    eapply pesim_real; intros ma ? RLa ?;
    destruct csa as [cpc ? ? ?], ma;
    destruct_tgt_real RLa;
    destruct cpc; [by vauto|clarify|by vauto];
    rrdes HR WHX WHX0; subst;
    hexploit tgt_confpred_shape; eauto;
    hexploit src_confpred_shape; eauto;
    hexploit ts_refconfrel_shape; eauto;
    hexploit mylw_confrel_shape; eauto;
    i; des; clarify; noemp;
    simpl cplus in *; clarify; noemp; clean;
    rewrite <-!pmplus_assoc in *;
    rrdes WHY; clarify; clean
  end.



Ltac ts_prepare :=
  let HR := fresh in
  intros ? ? HR; revert HR; ts_prepare'.


Lemma pesim_rewrite: forall La Lb (w: world La Lb)
    i kia kib g si s sync ty ea eb
    (X: esim w i kia kib g si s sync ty ea eb)
    i' (Ei: i = i')
    kia' (Ekia: kia = kia')
    kib' (Ekib: kib = kib')
    g' (Eg: g = g')
    si' (Esi: si = si')
    s' (Es: s = s')
    sync' (Esync: sync = sync')
    ty' (Ety: ty = ty')
    ea' (Eea: ea = ea')
    eb' (Eeb: eb = eb'),
  esim w i' kia' kib' g' si' s' sync' ty' ea' eb'.
Proof.
  i; subst; auto.
Qed.                            
